From 99d3ae9110e28020501a273accd5143464091af1 Mon Sep 17 00:00:00 2001
From: Kirithika <kirithika@multicorewareinc.com>
Date: Tue, 2 Jul 2024 15:03:03 +0530
Subject: [PATCH 04/20] Add encoding support for enhancement layer

---
 source/common/frame.h           |   3 +
 source/common/wavefront.cpp     |   7 +-
 source/common/wavefront.h       |   6 +-
 source/encoder/encoder.cpp      | 455 +++++++++++++++++---------------
 source/encoder/frameencoder.cpp | 319 +++++++++++-----------
 source/encoder/frameencoder.h   |  22 +-
 source/encoder/framefilter.cpp  |  12 +-
 source/encoder/framefilter.h    |   4 +-
 source/encoder/nal.cpp          |   4 +-
 source/encoder/nal.h            |   2 +-
 source/encoder/sei.cpp          |   4 +-
 source/encoder/sei.h            |   2 +-
 12 files changed, 441 insertions(+), 399 deletions(-)

diff --git a/source/common/frame.h b/source/common/frame.h
index 9a3a3a4f9..97094ad73 100644
--- a/source/common/frame.h
+++ b/source/common/frame.h
@@ -161,6 +161,9 @@ public:
     int8_t                 m_gopId;
     bool                   m_sameLayerRefPic;
 
+    int                    m_sLayerId;
+    bool                   m_valid;
+
     Frame();
 
     bool create(x265_param *param, float* quantOffsets);
diff --git a/source/common/wavefront.cpp b/source/common/wavefront.cpp
index 5b535aff5..c3f2d81a8 100644
--- a/source/common/wavefront.cpp
+++ b/source/common/wavefront.cpp
@@ -58,6 +58,11 @@ WaveFront::~WaveFront()
     x265_free((void*)m_externalDependencyBitmap);
 }
 
+void WaveFront::setLayerId(int layer)
+{
+    m_sLayerId = layer;
+}
+
 void WaveFront::clearEnabledRowMask()
 {
     memset((void*)m_externalDependencyBitmap, 0, sizeof(uint32_t) * m_numWords);
@@ -103,7 +108,7 @@ void WaveFront::findJob(int threadId)
             if (ATOMIC_AND(&m_internalDependencyBitmap[w], ~bit) & bit)
             {
                 /* we cleared the bit, we get to process the row */
-                processRow(w * 32 + id, threadId);
+                processRow(w * 32 + id, threadId, m_sLayerId);
                 m_helpWanted = true;
                 return; /* check for a higher priority task */
             }
diff --git a/source/common/wavefront.h b/source/common/wavefront.h
index 99d93c827..5f75d4b3b 100644
--- a/source/common/wavefront.h
+++ b/source/common/wavefront.h
@@ -52,6 +52,8 @@ private:
 
     int m_numRows;
 
+    int m_sLayerId;
+
 protected:
     uint32_t *m_row_to_idx;
     uint32_t *m_idx_to_row;
@@ -95,7 +97,9 @@ public:
 
     // Start or resume encode processing of this row, must be implemented by
     // derived classes.
-    virtual void processRow(int row, int threadId) = 0;
+    virtual void processRow(int row, int threadId, int layer) = 0;
+
+    void setLayerId(int layer);
 };
 } // end namespace X265_NS
 
diff --git a/source/encoder/encoder.cpp b/source/encoder/encoder.cpp
index ef6c5f64a..e62823ef6 100644
--- a/source/encoder/encoder.cpp
+++ b/source/encoder/encoder.cpp
@@ -1577,136 +1577,141 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
         else
             inputPic = pic_in;
 
-        Frame *inFrame;
+        Frame* inFrame[MAX_SCALABLE_LAYERS];
         x265_param *p = (m_reconfigure || m_reconfigureRc) ? m_latestParam : m_param;
-        if (m_dpb->m_freeList.empty())
+        for (int layer = 0; layer < MAX_SCALABLE_LAYERS; layer++)
         {
-            inFrame = new Frame;
-            inFrame->m_encodeStartTime = x265_mdate();
-            if (inFrame->create(p, inputPic->quantOffsets))
+            if (m_dpb->m_freeList.empty())
             {
-                /* the first PicYuv created is asked to generate the CU and block unit offset
-                 * arrays which are then shared with all subsequent PicYuv (orig and recon) 
-                 * allocated by this top level encoder */
-                if (m_sps.cuOffsetY)
+                inFrame[layer] = new Frame;
+                inFrame[layer]->m_encodeStartTime = x265_mdate();
+                inFrame[layer]->m_sLayerId = layer;
+                inFrame[layer]->m_valid = false;
+                if (inFrame[layer]->create(p, inputPic->quantOffsets))
                 {
-                    inFrame->m_fencPic->m_cuOffsetY = m_sps.cuOffsetY;
-                    inFrame->m_fencPic->m_buOffsetY = m_sps.buOffsetY;
-                    if (m_param->internalCsp != X265_CSP_I400)
+                    /* the first PicYuv created is asked to generate the CU and block unit offset
+                     * arrays which are then shared with all subsequent PicYuv (orig and recon)
+                     * allocated by this top level encoder */
+                    if (m_sps.cuOffsetY)
                     {
-                        inFrame->m_fencPic->m_cuOffsetC = m_sps.cuOffsetC;
-                        inFrame->m_fencPic->m_buOffsetC = m_sps.buOffsetC;
-                    }
-                }
-                else
-                {
-                    if (!inFrame->m_fencPic->createOffsets(m_sps))
-                    {
-                        m_aborted = true;
-                        x265_log(m_param, X265_LOG_ERROR, "memory allocation failure, aborting encode\n");
-                        inFrame->destroy();
-                        delete inFrame;
-                        return -1;
+                        inFrame[layer]->m_fencPic->m_cuOffsetY = m_sps.cuOffsetY;
+                        inFrame[layer]->m_fencPic->m_buOffsetY = m_sps.buOffsetY;
+                        if (m_param->internalCsp != X265_CSP_I400)
+                        {
+                            inFrame[layer]->m_fencPic->m_cuOffsetC = m_sps.cuOffsetC;
+                            inFrame[layer]->m_fencPic->m_buOffsetC = m_sps.buOffsetC;
+                        }
                     }
                     else
                     {
-                        m_sps.cuOffsetY = inFrame->m_fencPic->m_cuOffsetY;
-                        m_sps.buOffsetY = inFrame->m_fencPic->m_buOffsetY;
-                        if (m_param->internalCsp != X265_CSP_I400)
+                        if (!inFrame[layer]->m_fencPic->createOffsets(m_sps))
                         {
-                            m_sps.cuOffsetC = inFrame->m_fencPic->m_cuOffsetC;
-                            m_sps.cuOffsetY = inFrame->m_fencPic->m_cuOffsetY;
-                            m_sps.buOffsetC = inFrame->m_fencPic->m_buOffsetC;
-                            m_sps.buOffsetY = inFrame->m_fencPic->m_buOffsetY;
+                            m_aborted = true;
+                            x265_log(m_param, X265_LOG_ERROR, "memory allocation failure, aborting encode\n");
+                            inFrame[layer]->destroy();
+                            delete inFrame[layer];
+                            return -1;
+                        }
+                        else
+                        {
+                            m_sps.cuOffsetY = inFrame[layer]->m_fencPic->m_cuOffsetY;
+                            m_sps.buOffsetY = inFrame[layer]->m_fencPic->m_buOffsetY;
+                            if (m_param->internalCsp != X265_CSP_I400)
+                            {
+                                m_sps.cuOffsetC = inFrame[layer]->m_fencPic->m_cuOffsetC;
+                                m_sps.cuOffsetY = inFrame[layer]->m_fencPic->m_cuOffsetY;
+                                m_sps.buOffsetC = inFrame[layer]->m_fencPic->m_buOffsetC;
+                                m_sps.buOffsetY = inFrame[layer]->m_fencPic->m_buOffsetY;
+                            }
                         }
                     }
                 }
+                else
+                {
+                    m_aborted = true;
+                    x265_log(m_param, X265_LOG_ERROR, "memory allocation failure, aborting encode\n");
+                    inFrame[layer]->destroy();
+                    delete inFrame[layer];
+                    return -1;
+                }
             }
             else
             {
-                m_aborted = true;
-                x265_log(m_param, X265_LOG_ERROR, "memory allocation failure, aborting encode\n");
-                inFrame->destroy();
-                delete inFrame;
-                return -1;
+                inFrame[layer] = m_dpb->m_freeList.popBack();
+                inFrame[layer]->m_encodeStartTime = x265_mdate();
+                /* Set lowres scencut and satdCost here to aovid overwriting ANALYSIS_READ
+                   decision by lowres init*/
+                inFrame[layer]->m_lowres.bScenecut = false;
+                inFrame[layer]->m_lowres.satdCost = (int64_t)-1;
+                inFrame[layer]->m_lowresInit = false;
+                inFrame[layer]->m_isInsideWindow = 0;
+                inFrame[layer]->m_tempLayer = 0;
+                inFrame[layer]->m_sameLayerRefPic = 0;
+                inFrame[layer]->m_sLayerId = layer;
             }
-        }
-        else
-        {
-            inFrame = m_dpb->m_freeList.popBack();
-            inFrame->m_encodeStartTime = x265_mdate();
-            /* Set lowres scencut and satdCost here to aovid overwriting ANALYSIS_READ
-               decision by lowres init*/
-            inFrame->m_lowres.bScenecut = false;
-            inFrame->m_lowres.satdCost = (int64_t)-1;
-            inFrame->m_lowresInit = false;
-            inFrame->m_isInsideWindow = 0;
-            inFrame->m_tempLayer = 0;
-            inFrame->m_sameLayerRefPic = 0;
-        }
 
-        /* Copy input picture into a Frame and PicYuv, send to lookahead */
-        inFrame->m_fencPic->copyFromPicture(*inputPic, *m_param, m_sps.conformanceWindow.rightOffset, m_sps.conformanceWindow.bottomOffset, true);
+            /* Copy input picture into a Frame and PicYuv, send to lookahead */
+            inFrame[layer]->m_fencPic->copyFromPicture(*inputPic, *m_param, m_sps.conformanceWindow.rightOffset, m_sps.conformanceWindow.bottomOffset, !layer);
 
-        inFrame->m_poc       = ++m_pocLast;
-        inFrame->m_userData  = inputPic->userData;
-        inFrame->m_pts       = inputPic->pts;
+            inFrame[layer]->m_poc = (!layer) ? (++m_pocLast) : m_pocLast;
+            inFrame[layer]->m_userData = inputPic->userData;
+            inFrame[layer]->m_pts = inputPic->pts;
 
-        if ((m_param->bEnableSceneCutAwareQp & BACKWARD) && m_param->rc.bStatRead)
-        {
-            RateControlEntry * rcEntry = NULL;
-            rcEntry = &(m_rateControl->m_rce2Pass[inFrame->m_poc]);
-            if(rcEntry->scenecut)
+            if ((m_param->bEnableSceneCutAwareQp & BACKWARD) && m_param->rc.bStatRead)
             {
-                int backwardWindow = X265_MIN(int((m_param->bwdMaxScenecutWindow / 1000.0) * (m_param->fpsNum / m_param->fpsDenom)), p->lookaheadDepth);
-                for (int i = 1; i <= backwardWindow; i++)
+                RateControlEntry* rcEntry = NULL;
+                rcEntry = &(m_rateControl->m_rce2Pass[inFrame[layer]->m_poc]);
+                if (rcEntry->scenecut)
                 {
-                    int frameNum = inFrame->m_poc - i;
-                    Frame * frame = m_lookahead->m_inputQueue.getPOC(frameNum);
-                    if (frame)
-                        frame->m_isInsideWindow = BACKWARD_WINDOW;
+                    int backwardWindow = X265_MIN(int((m_param->bwdMaxScenecutWindow / 1000.0) * (m_param->fpsNum / m_param->fpsDenom)), p->lookaheadDepth);
+                    for (int i = 1; i <= backwardWindow; i++)
+                    {
+                        int frameNum = inFrame[layer]->m_poc - i;
+                        Frame* frame = m_lookahead->m_inputQueue.getPOC(frameNum);
+                        if (frame)
+                            frame->m_isInsideWindow = BACKWARD_WINDOW;
+                    }
                 }
             }
-        }
 
-        inFrame->m_forceqp   = inputPic->forceqp;
-        inFrame->m_param     = (m_reconfigure || m_reconfigureRc) ? m_latestParam : m_param;
-        inFrame->m_picStruct = inputPic->picStruct;
-        if (m_param->bField && m_param->interlaceMode)
-            inFrame->m_fieldNum = inputPic->fieldNum;
+            inFrame[layer]->m_forceqp = inputPic->forceqp;
+            inFrame[layer]->m_param = (m_reconfigure || m_reconfigureRc) ? m_latestParam : m_param;
+            inFrame[layer]->m_picStruct = inputPic->picStruct;
+            if (m_param->bField && m_param->interlaceMode)
+                inFrame[layer]->m_fieldNum = inputPic->fieldNum;
 
-        copyUserSEIMessages(inFrame, inputPic);
+            /* Encoder holds a reference count until stats collection is finished */
+            ATOMIC_INC(&inFrame[layer]->m_countRefEncoders);
+        }
+        copyUserSEIMessages(inFrame[0], inputPic);
 
         /*Copy Dolby Vision RPU from inputPic to frame*/
         if (inputPic->rpu.payloadSize)
         {
-            inFrame->m_rpu.payloadSize = inputPic->rpu.payloadSize;
-            inFrame->m_rpu.payload = new uint8_t[inputPic->rpu.payloadSize];
-            memcpy(inFrame->m_rpu.payload, inputPic->rpu.payload, inputPic->rpu.payloadSize);
+            inFrame[0]->m_rpu.payloadSize = inputPic->rpu.payloadSize;
+            inFrame[0]->m_rpu.payload = new uint8_t[inputPic->rpu.payloadSize];
+            memcpy(inFrame[0]->m_rpu.payload, inputPic->rpu.payload, inputPic->rpu.payloadSize);
         }
 
         if (inputPic->quantOffsets != NULL)
         {
             int cuCount;
             if (m_param->rc.qgSize == 8)
-                cuCount = inFrame->m_lowres.maxBlocksInRowFullRes * inFrame->m_lowres.maxBlocksInColFullRes;
+                cuCount = inFrame[0]->m_lowres.maxBlocksInRowFullRes * inFrame[0]->m_lowres.maxBlocksInColFullRes;
             else
-                cuCount = inFrame->m_lowres.maxBlocksInRow * inFrame->m_lowres.maxBlocksInCol;
-            memcpy(inFrame->m_quantOffsets, inputPic->quantOffsets, cuCount * sizeof(float));
+                cuCount = inFrame[0]->m_lowres.maxBlocksInRow * inFrame[0]->m_lowres.maxBlocksInCol;
+            memcpy(inFrame[0]->m_quantOffsets, inputPic->quantOffsets, cuCount * sizeof(float));
         }
 
         if (m_pocLast == 0)
-            m_firstPts = inFrame->m_pts;
+            m_firstPts = inFrame[0]->m_pts;
         if (m_bframeDelay && m_pocLast == m_bframeDelay)
-            m_bframeDelayTime = inFrame->m_pts - m_firstPts;
-
-        /* Encoder holds a reference count until stats collection is finished */
-        ATOMIC_INC(&inFrame->m_countRefEncoders);
+            m_bframeDelayTime = inFrame[0]->m_pts - m_firstPts;
 
         if ((m_param->rc.aqMode || m_param->bEnableWeightedPred || m_param->bEnableWeightedBiPred) &&
             (m_param->rc.cuTree && m_param->rc.bStatRead))
         {
-            if (!m_rateControl->cuTreeReadFor2Pass(inFrame))
+            if (!m_rateControl->cuTreeReadFor2Pass(inFrame[0]))
             {
                 m_aborted = 1;
                 return -1;
@@ -1714,8 +1719,8 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
         }
 
         /* Use the frame types from the first pass, if available */
-        int sliceType = (m_param->rc.bStatRead) ? m_rateControl->rateControlSliceType(inFrame->m_poc) : X265_TYPE_AUTO;
-        inFrame->m_lowres.sliceTypeReq = inputPic->sliceType;
+        int sliceType = (m_param->rc.bStatRead) ? m_rateControl->rateControlSliceType(inFrame[0]->m_poc) : X265_TYPE_AUTO;
+        inFrame[0]->m_lowres.sliceTypeReq = inputPic->sliceType;
 
         /* In analysisSave mode, x265_analysis_data is allocated in inputPic and inFrame points to this */
         /* Load analysis data before lookahead->addPicture, since sliceType has been decided */
@@ -1723,7 +1728,7 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
         {
             /* reads analysis data for the frame and allocates memory based on slicetype */
             static int paramBytes = CONF_OFFSET_BYTES;
-            if (!inFrame->m_poc && m_param->bAnalysisType != HEVC_INFO)
+            if (!inFrame[0]->m_poc && m_param->bAnalysisType != HEVC_INFO)
             {
                 x265_analysis_validate saveParam = inputPic->analysisData.saveParam;
                 paramBytes += validateAnalysisData(&saveParam, 0);
@@ -1746,26 +1751,26 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
                 uint32_t outOfBoundaryLowresH = extendedHeight - m_param->sourceHeight / 2;
                 if (outOfBoundaryLowresH * 2 >= m_param->maxCUSize)
                     cuLocInFrame.skipHeight = true;
-                readAnalysisFile(&inFrame->m_analysisData, inFrame->m_poc, inputPic, paramBytes, cuLocInFrame);
+                readAnalysisFile(&inFrame[0]->m_analysisData, inFrame[0]->m_poc, inputPic, paramBytes, cuLocInFrame);
             }
             else
-                readAnalysisFile(&inFrame->m_analysisData, inFrame->m_poc, inputPic, paramBytes);
-            inFrame->m_poc = inFrame->m_analysisData.poc;
-            sliceType = inFrame->m_analysisData.sliceType;
-            inFrame->m_lowres.bScenecut = !!inFrame->m_analysisData.bScenecut;
-            inFrame->m_lowres.satdCost = inFrame->m_analysisData.satdCost;
+                readAnalysisFile(&inFrame[0]->m_analysisData, inFrame[0]->m_poc, inputPic, paramBytes);
+            inFrame[0]->m_poc = inFrame[0]->m_analysisData.poc;
+            sliceType = inFrame[0]->m_analysisData.sliceType;
+            inFrame[0]->m_lowres.bScenecut = !!inFrame[0]->m_analysisData.bScenecut;
+            inFrame[0]->m_lowres.satdCost = inFrame[0]->m_analysisData.satdCost;
             if (m_param->bDisableLookahead)
             {
-                inFrame->m_lowres.sliceType = sliceType;
-                inFrame->m_lowres.bKeyframe = !!inFrame->m_analysisData.lookahead.keyframe;
-                inFrame->m_lowres.bLastMiniGopBFrame = !!inFrame->m_analysisData.lookahead.lastMiniGopBFrame;
+                inFrame[0]->m_lowres.sliceType = sliceType;
+                inFrame[0]->m_lowres.bKeyframe = !!inFrame[0]->m_analysisData.lookahead.keyframe;
+                inFrame[0]->m_lowres.bLastMiniGopBFrame = !!inFrame[0]->m_analysisData.lookahead.lastMiniGopBFrame;
                 if (m_rateControl->m_isVbv)
                 {
                     int vbvCount = m_param->lookaheadDepth + m_param->bframes + 2;
                     for (int index = 0; index < vbvCount; index++)
                     {
-                        inFrame->m_lowres.plannedSatd[index] = inFrame->m_analysisData.lookahead.plannedSatd[index];
-                        inFrame->m_lowres.plannedType[index] = inFrame->m_analysisData.lookahead.plannedType[index];
+                        inFrame[0]->m_lowres.plannedSatd[index] = inFrame[0]->m_analysisData.lookahead.plannedSatd[index];
+                        inFrame[0]->m_lowres.plannedType[index] = inFrame[0]->m_analysisData.lookahead.plannedType[index];
                     }
                 }
             }
@@ -1809,14 +1814,14 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
         }
 
         if (m_reconfigureRc)
-            inFrame->m_reconfigureRc = true;
+            inFrame[0]->m_reconfigureRc = true;
 
         if (m_param->bEnableTemporalFilter)
         {
             if (!m_pocLast)
             {
                 /*One shot allocation of frames in OriginalPictureBuffer*/
-                int numFramesinOPB = X265_MAX(m_param->bframes, (inFrame->m_mcstf->m_range << 1)) + 1;
+                int numFramesinOPB = X265_MAX(m_param->bframes, (inFrame[0]->m_mcstf->m_range << 1)) + 1;
                 for (int i = 0; i < numFramesinOPB; i++)
                 {
                     Frame* dupFrame = new Frame;
@@ -1848,23 +1853,25 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
                 }
             }
 
-            inFrame->m_refPicCnt[1] = 2 * inFrame->m_mcstf->m_range + 1;
-            if (inFrame->m_poc < inFrame->m_mcstf->m_range)
-                inFrame->m_refPicCnt[1] -= (uint8_t)(inFrame->m_mcstf->m_range - inFrame->m_poc);
-            if (m_param->totalFrames && (inFrame->m_poc >= (m_param->totalFrames - inFrame->m_mcstf->m_range)))
-                inFrame->m_refPicCnt[1] -= (uint8_t)(inFrame->m_poc + inFrame->m_mcstf->m_range - m_param->totalFrames + 1);
+            inFrame[0]->m_refPicCnt[1] = 2 * inFrame[0]->m_mcstf->m_range + 1;
+            if (inFrame[0]->m_poc < inFrame[0]->m_mcstf->m_range)
+                inFrame[0]->m_refPicCnt[1] -= (uint8_t)(inFrame[0]->m_mcstf->m_range - inFrame[0]->m_poc);
+            if (m_param->totalFrames && (inFrame[0]->m_poc >= (m_param->totalFrames - inFrame[0]->m_mcstf->m_range)))
+                inFrame[0]->m_refPicCnt[1] -= (uint8_t)(inFrame[0]->m_poc + inFrame[0]->m_mcstf->m_range - m_param->totalFrames + 1);
 
             //Extend full-res original picture border
-            PicYuv *orig = inFrame->m_fencPic;
+            PicYuv *orig = inFrame[0]->m_fencPic;
             extendPicBorder(orig->m_picOrg[0], orig->m_stride, orig->m_picWidth, orig->m_picHeight, orig->m_lumaMarginX, orig->m_lumaMarginY);
             extendPicBorder(orig->m_picOrg[1], orig->m_strideC, orig->m_picWidth >> orig->m_hChromaShift, orig->m_picHeight >> orig->m_vChromaShift, orig->m_chromaMarginX, orig->m_chromaMarginY);
             extendPicBorder(orig->m_picOrg[2], orig->m_strideC, orig->m_picWidth >> orig->m_hChromaShift, orig->m_picHeight >> orig->m_vChromaShift, orig->m_chromaMarginX, orig->m_chromaMarginY);
 
             //TODO: Add subsampling here if required
-            m_origPicBuffer->addPicture(inFrame);
+            m_origPicBuffer->addPicture(inFrame[0]);
         }
 
-        m_lookahead->addPicture(*inFrame, sliceType);
+        m_lookahead->addPicture(*inFrame[0], sliceType);
+
+        m_dpb->m_picList.pushBack(*inFrame[1]); /* Add enhancement layer to DPB to be used later in frameencoder*/
         m_numDelayedPic++;
     }
     else if (m_latestParam->forceFlush == 2)
@@ -1881,7 +1888,7 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
      * input picture before returning so the order must be reversed. This do/while() loop allows
      * us to alternate the order of the calls without ugly code replication */
     Frame* outFrame = NULL;
-    Frame* frameEnc = NULL;
+    Frame* frameEnc[MAX_SCALABLE_LAYERS] = { NULL };
     int pass = 0;
     do
     {
@@ -2131,23 +2138,28 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
         /* pop a single frame from decided list, then provide to frame encoder
          * curEncoder is guaranteed to be idle at this point */
         if (!pass)
-            frameEnc = m_lookahead->getDecidedPicture();
-        if (frameEnc && !pass && (!m_param->chunkEnd || (m_encodedFrameNum < m_param->chunkEnd)))
+            frameEnc[0] = m_lookahead->getDecidedPicture();
+        if (frameEnc[0] && !pass && (!m_param->chunkEnd || (m_encodedFrameNum < m_param->chunkEnd)))
         {
+            //Pop non base view pictures from DPB piclist
+            frameEnc[1] = m_dpb->m_picList.getPOC(frameEnc[0]->m_poc);
+            m_dpb->m_picList.remove(*frameEnc[1]);
+            frameEnc[1]->m_lowres.sliceType = frameEnc[0]->m_lowres.sliceType;
+
             if ((m_param->bEnableSceneCutAwareQp & FORWARD) && m_param->rc.bStatRead)
             {
                 RateControlEntry * rcEntry;
-                rcEntry = &(m_rateControl->m_rce2Pass[frameEnc->m_poc]);
+                rcEntry = &(m_rateControl->m_rce2Pass[frameEnc[0]->m_poc]);
 
                 if (rcEntry->scenecut)
                 {
                     if (m_rateControl->m_lastScenecut == -1)
-                        m_rateControl->m_lastScenecut = frameEnc->m_poc;
+                        m_rateControl->m_lastScenecut = frameEnc[0]->m_poc;
                     else
                     {
                         int maxWindowSize = int((m_param->fwdMaxScenecutWindow / 1000.0) * (m_param->fpsNum / m_param->fpsDenom) + 0.5);
-                        if (frameEnc->m_poc > (m_rateControl->m_lastScenecut + maxWindowSize))
-                            m_rateControl->m_lastScenecut = frameEnc->m_poc;
+                        if (frameEnc[0]->m_poc > (m_rateControl->m_lastScenecut + maxWindowSize))
+                            m_rateControl->m_lastScenecut = frameEnc[0]->m_poc;
                     }
                 }
             }
@@ -2156,31 +2168,31 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
             {
                 uint32_t widthInCU = (m_param->sourceWidth + m_param->maxCUSize - 1) >> m_param->maxLog2CUSize;
                 uint32_t heightInCU = (m_param->sourceHeight + m_param->maxCUSize - 1) >> m_param->maxLog2CUSize;
-                frameEnc->m_analysisData.numCUsInFrame = widthInCU * heightInCU;
-                frameEnc->m_analysisData.numPartitions = m_param->num4x4Partitions;
-                x265_alloc_analysis_data(m_param, &frameEnc->m_analysisData);
-                frameEnc->m_analysisData.poc = frameEnc->m_poc;
+                frameEnc[0]->m_analysisData.numCUsInFrame = widthInCU * heightInCU;
+                frameEnc[0]->m_analysisData.numPartitions = m_param->num4x4Partitions;
+                x265_alloc_analysis_data(m_param, &frameEnc[0]->m_analysisData);
+                frameEnc[0]->m_analysisData.poc = frameEnc[0]->m_poc;
                 if (m_param->rc.bStatRead)
-                    readAnalysisFile(&frameEnc->m_analysisData, frameEnc->m_poc, frameEnc->m_lowres.sliceType);
+                    readAnalysisFile(&frameEnc[0]->m_analysisData, frameEnc[0]->m_poc, frameEnc[0]->m_lowres.sliceType);
             }
 
             if (m_param->bResetZoneConfig)
             {
                 for (int i = 0; i < m_param->rc.zonefileCount; i++)
                 {
-                    if (m_param->rc.zones[i].startFrame == frameEnc->m_poc)
+                    if (m_param->rc.zones[i].startFrame == frameEnc[0]->m_poc)
                         x265_encoder_reconfig(this, m_param->rc.zones[i].zoneParam);
                 }
             }
 
-            if (frameEnc->m_reconfigureRc && m_reconfigureRc)
+            if (frameEnc[0]->m_reconfigureRc && m_reconfigureRc)
             {
                 x265_copy_params(m_param, m_latestParam);
                 m_rateControl->reconfigureRC();
                 m_reconfigureRc = false;
             }
-            if (frameEnc->m_reconfigureRc && !m_reconfigureRc)
-                frameEnc->m_reconfigureRc = false;
+            if (frameEnc[0]->m_reconfigureRc && !m_reconfigureRc)
+                frameEnc[0]->m_reconfigureRc = false;
             if (curEncoder->m_reconfigure)
             {
                 /* One round robin cycle of FE reconfigure is complete */
@@ -2196,62 +2208,66 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
             curEncoder->m_reconfigure = m_reconfigure;
 
             /* give this frame a FrameData instance before encoding */
-            if (m_dpb->m_frameDataFreeList)
+            for (int layer = 0; layer < MAX_SCALABLE_LAYERS; layer++)
             {
-                frameEnc->m_encData = m_dpb->m_frameDataFreeList;
-                m_dpb->m_frameDataFreeList = m_dpb->m_frameDataFreeList->m_freeListNext;
-                frameEnc->reinit(m_sps);
-                frameEnc->m_param = m_reconfigure ? m_latestParam : m_param;
-                frameEnc->m_encData->m_param = m_reconfigure ? m_latestParam : m_param;
-            }
-            else
-            {
-                frameEnc->allocEncodeData(m_reconfigure ? m_latestParam : m_param, m_sps);
-                Slice* slice = frameEnc->m_encData->m_slice;
-                slice->m_sps = &m_sps;
-                slice->m_pps = &m_pps;
-                slice->m_param = m_param;
-                slice->m_maxNumMergeCand = m_param->maxNumMergeCand;
-                slice->m_endCUAddr = slice->realEndAddress(m_sps.numCUsInFrame * m_param->num4x4Partitions);
+                if (m_dpb->m_frameDataFreeList)
+                {
+                    frameEnc[layer]->m_encData = m_dpb->m_frameDataFreeList;
+                    m_dpb->m_frameDataFreeList = m_dpb->m_frameDataFreeList->m_freeListNext;
+                    frameEnc[layer]->reinit(m_sps);
+                    frameEnc[layer]->m_param = m_reconfigure ? m_latestParam : m_param;
+                    frameEnc[layer]->m_encData->m_param = m_reconfigure ? m_latestParam : m_param;
+                }
+                else
+                {
+                    frameEnc[layer]->allocEncodeData(m_reconfigure ? m_latestParam : m_param, m_sps);
+                    Slice* slice = frameEnc[layer]->m_encData->m_slice;
+                    slice->m_sps = &m_sps;
+                    slice->m_pps = &m_pps;
+                    slice->m_param = m_param;
+                    slice->m_maxNumMergeCand = m_param->maxNumMergeCand;
+                    slice->m_endCUAddr = slice->realEndAddress(m_sps.numCUsInFrame * m_param->num4x4Partitions);
+                }
+                frameEnc[layer]->m_valid = true;
             }
             if (m_param->analysisLoad && m_param->bDisableLookahead)
             {
-                frameEnc->m_dts = frameEnc->m_analysisData.lookahead.dts;
-                frameEnc->m_reorderedPts = frameEnc->m_analysisData.lookahead.reorderedPts;
+                frameEnc[0]->m_dts = frameEnc[0]->m_analysisData.lookahead.dts;
+                frameEnc[0]->m_reorderedPts = frameEnc[0]->m_analysisData.lookahead.reorderedPts;
                 if (m_rateControl->m_isVbv)
                 {
-                    for (uint32_t index = 0; index < frameEnc->m_analysisData.numCuInHeight; index++)
+                    for (uint32_t index = 0; index < frameEnc[0]->m_analysisData.numCuInHeight; index++)
                     {
-                        frameEnc->m_encData->m_rowStat[index].intraSatdForVbv = frameEnc->m_analysisData.lookahead.intraSatdForVbv[index];
-                        frameEnc->m_encData->m_rowStat[index].satdForVbv = frameEnc->m_analysisData.lookahead.satdForVbv[index];
+                        frameEnc[0]->m_encData->m_rowStat[index].intraSatdForVbv = frameEnc[0]->m_analysisData.lookahead.intraSatdForVbv[index];
+                        frameEnc[0]->m_encData->m_rowStat[index].satdForVbv = frameEnc[0]->m_analysisData.lookahead.satdForVbv[index];
                     }
-                    for (uint32_t index = 0; index < frameEnc->m_analysisData.numCUsInFrame; index++)
+                    for (uint32_t index = 0; index < frameEnc[0]->m_analysisData.numCUsInFrame; index++)
                     {
-                        frameEnc->m_encData->m_cuStat[index].intraVbvCost = frameEnc->m_analysisData.lookahead.intraVbvCost[index];
-                        frameEnc->m_encData->m_cuStat[index].vbvCost = frameEnc->m_analysisData.lookahead.vbvCost[index];
+                        frameEnc[0]->m_encData->m_cuStat[index].intraVbvCost = frameEnc[0]->m_analysisData.lookahead.intraVbvCost[index];
+                        frameEnc[0]->m_encData->m_cuStat[index].vbvCost = frameEnc[0]->m_analysisData.lookahead.vbvCost[index];
                     }
                 }
             }
-            if (m_param->searchMethod == X265_SEA && frameEnc->m_lowres.sliceType != X265_TYPE_B)
+            if (m_param->searchMethod == X265_SEA && frameEnc[0]->m_lowres.sliceType != X265_TYPE_B)
             {
                 int padX = m_param->maxCUSize + 32;
                 int padY = m_param->maxCUSize + 16;
-                uint32_t numCuInHeight = (frameEnc->m_encData->m_reconPic->m_picHeight + m_param->maxCUSize - 1) / m_param->maxCUSize;
+                uint32_t numCuInHeight = (frameEnc[0]->m_encData->m_reconPic->m_picHeight + m_param->maxCUSize - 1) / m_param->maxCUSize;
                 int maxHeight = numCuInHeight * m_param->maxCUSize;
                 for (int i = 0; i < INTEGRAL_PLANE_NUM; i++)
                 {
-                    frameEnc->m_encData->m_meBuffer[i] = X265_MALLOC(uint32_t, frameEnc->m_reconPic->m_stride * (maxHeight + (2 * padY)));
-                    if (frameEnc->m_encData->m_meBuffer[i])
+                    frameEnc[0]->m_encData->m_meBuffer[i] = X265_MALLOC(uint32_t, frameEnc[0]->m_reconPic->m_stride * (maxHeight + (2 * padY)));
+                    if (frameEnc[0]->m_encData->m_meBuffer[i])
                     {
-                        memset(frameEnc->m_encData->m_meBuffer[i], 0, sizeof(uint32_t)* frameEnc->m_reconPic->m_stride * (maxHeight + (2 * padY)));
-                        frameEnc->m_encData->m_meIntegral[i] = frameEnc->m_encData->m_meBuffer[i] + frameEnc->m_encData->m_reconPic->m_stride * padY + padX;
+                        memset(frameEnc[0]->m_encData->m_meBuffer[i], 0, sizeof(uint32_t)* frameEnc[0]->m_reconPic->m_stride * (maxHeight + (2 * padY)));
+                        frameEnc[0]->m_encData->m_meIntegral[i] = frameEnc[0]->m_encData->m_meBuffer[i] + frameEnc[0]->m_encData->m_reconPic->m_stride * padY + padX;
                     }
                     else
-                        x265_log(m_param, X265_LOG_ERROR, "SEA motion search: POC %d Integral buffer[%d] unallocated\n", frameEnc->m_poc, i);
+                        x265_log(m_param, X265_LOG_ERROR, "SEA motion search: POC %d Integral buffer[%d] unallocated\n", frameEnc[0]->m_poc, i);
                 }
             }
 
-            if (m_param->bOptQpPPS && frameEnc->m_lowres.bKeyframe && m_param->bRepeatHeaders)
+            if (m_param->bOptQpPPS && frameEnc[0]->m_lowres.bKeyframe && m_param->bRepeatHeaders)
             {
                 ScopedLock qpLock(m_sliceQpLock);
                 if (m_iFrameNum > 0)
@@ -2278,33 +2294,33 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
                     m_iBitsCostSum[i] = 0;
             }
 
-            frameEnc->m_encData->m_slice->m_iPPSQpMinus26 = m_iPPSQpMinus26;
-            frameEnc->m_encData->m_slice->numRefIdxDefault[0] = m_pps.numRefIdxDefault[0];
-            frameEnc->m_encData->m_slice->numRefIdxDefault[1] = m_pps.numRefIdxDefault[1];
-            frameEnc->m_encData->m_slice->m_iNumRPSInSPS = m_sps.spsrpsNum;
+            frameEnc[0]->m_encData->m_slice->m_iPPSQpMinus26 = m_iPPSQpMinus26;
+            frameEnc[0]->m_encData->m_slice->numRefIdxDefault[0] = m_pps.numRefIdxDefault[0];
+            frameEnc[0]->m_encData->m_slice->numRefIdxDefault[1] = m_pps.numRefIdxDefault[1];
+            frameEnc[0]->m_encData->m_slice->m_iNumRPSInSPS = m_sps.spsrpsNum;
 
-            curEncoder->m_rce.encodeOrder = frameEnc->m_encodeOrder = m_encodedFrameNum++;
+            curEncoder->m_rce.encodeOrder = frameEnc[0]->m_encodeOrder = m_encodedFrameNum++;
 
             if (!m_param->analysisLoad || !m_param->bDisableLookahead)
             {
                 if (m_bframeDelay)
                 {
                     int64_t *prevReorderedPts = m_prevReorderedPts;
-                    frameEnc->m_dts = m_encodedFrameNum > m_bframeDelay
+                    frameEnc[0]->m_dts = m_encodedFrameNum > m_bframeDelay
                         ? prevReorderedPts[(m_encodedFrameNum - m_bframeDelay) % m_bframeDelay]
-                        : frameEnc->m_reorderedPts - m_bframeDelayTime;
-                    prevReorderedPts[m_encodedFrameNum % m_bframeDelay] = frameEnc->m_reorderedPts;
+                        : frameEnc[0]->m_reorderedPts - m_bframeDelayTime;
+                    prevReorderedPts[m_encodedFrameNum % m_bframeDelay] = frameEnc[0]->m_reorderedPts;
                 }
                 else
-                    frameEnc->m_dts = frameEnc->m_reorderedPts;
+                    frameEnc[0]->m_dts = frameEnc[0]->m_reorderedPts;
             }
 
-            /* Allocate analysis data before encode in save mode. This is allocated in frameEnc */
+            /* Allocate analysis data before encode in save mode. This is allocated in frameEnc[0] */
             if (m_param->analysisSave && !m_param->analysisLoad)
             {
-                x265_analysis_data* analysis = &frameEnc->m_analysisData;
-                analysis->poc = frameEnc->m_poc;
-                analysis->sliceType = frameEnc->m_lowres.sliceType;
+                x265_analysis_data* analysis = &frameEnc[0]->m_analysisData;
+                analysis->poc = frameEnc[0]->m_poc;
+                analysis->sliceType = frameEnc[0]->m_lowres.sliceType;
                 uint32_t widthInCU       = (m_param->sourceWidth  + m_param->maxCUSize - 1) >> m_param->maxLog2CUSize;
                 uint32_t heightInCU      = (m_param->sourceHeight + m_param->maxCUSize - 1) >> m_param->maxLog2CUSize;
 
@@ -2317,22 +2333,23 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
             if (m_param->bEnableTemporalSubLayers > 2)
             {
                 //Re-assign temporalid if the current frame is at the end of encode or when I slice is encountered
-                if ((frameEnc->m_poc == (m_param->totalFrames - 1)) || (frameEnc->m_lowres.sliceType == X265_TYPE_I) || (frameEnc->m_lowres.sliceType == X265_TYPE_IDR))
+                if ((frameEnc[0]->m_poc == (m_param->totalFrames - 1)) || (frameEnc[0]->m_lowres.sliceType == X265_TYPE_I) || (frameEnc[0]->m_lowres.sliceType == X265_TYPE_IDR))
                 {
-                    frameEnc->m_tempLayer = (int8_t)0;
+                    frameEnc[0]->m_tempLayer = (int8_t)0;
                 }
             }
             /* determine references, setup RPS, etc */
-            m_dpb->prepareEncode(frameEnc);
+            for (int layer = 0; layer < MAX_SCALABLE_LAYERS; layer++)
+                m_dpb->prepareEncode(frameEnc[layer]);
 
             if (m_param->bEnableTemporalFilter)
             {
                 X265_CHECK(!m_origPicBuffer->m_mcstfOrigPicFreeList.empty(), "Frames not available in Encoded OPB");
 
                 Frame *dupFrame = m_origPicBuffer->m_mcstfOrigPicFreeList.popBackMCSTF();
-                dupFrame->m_fencPic->copyFromFrame(frameEnc->m_fencPic);
-                dupFrame->m_poc = frameEnc->m_poc;
-                dupFrame->m_encodeOrder = frameEnc->m_encodeOrder;
+                dupFrame->m_fencPic->copyFromFrame(frameEnc[0]->m_fencPic);
+                dupFrame->m_poc = frameEnc[0]->m_poc;
+                dupFrame->m_encodeOrder = frameEnc[0]->m_encodeOrder;
                 dupFrame->m_refPicCnt[1] = 2 * dupFrame->m_mcstf->m_range + 1;
 
                 if (dupFrame->m_poc < dupFrame->m_mcstf->m_range)
@@ -2341,61 +2358,63 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
                     dupFrame->m_refPicCnt[1] -= (uint8_t)(dupFrame->m_poc + dupFrame->m_mcstf->m_range - m_param->totalFrames + 1);
 
                 m_origPicBuffer->addEncPictureToPicList(dupFrame);
-                m_origPicBuffer->setOrigPicList(frameEnc, m_pocLast);
+                m_origPicBuffer->setOrigPicList(frameEnc[0], m_pocLast);
             }
 
-            if (!!m_param->selectiveSAO)
+            for (int layer = 0; layer < MAX_SCALABLE_LAYERS; layer++)
             {
-                Slice* slice = frameEnc->m_encData->m_slice;
-                slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 1;
-                switch (m_param->selectiveSAO)
+                if (!!m_param->selectiveSAO)
                 {
-                case 3: if (!IS_REFERENCED(frameEnc))
-                            slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
+                    Slice* slice = frameEnc[layer]->m_encData->m_slice;
+                    slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 1;
+                    switch (m_param->selectiveSAO)
+                    {
+                    case 3: if (!IS_REFERENCED(frameEnc[layer]))
+                        slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
                         break;
-                case 2: if (!!m_param->bframes && slice->m_sliceType == B_SLICE)
-                            slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
+                    case 2: if (!!m_param->bframes && slice->m_sliceType == B_SLICE)
+                        slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
                         break;
-                case 1: if (slice->m_sliceType != I_SLICE)
-                            slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
+                    case 1: if (slice->m_sliceType != I_SLICE)
+                        slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
                         break;
+                    }
+                }
+                else
+                {
+                    Slice* slice = frameEnc[layer]->m_encData->m_slice;
+                    slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
                 }
             }
-            else
-            {
-                Slice* slice = frameEnc->m_encData->m_slice;
-                slice->m_bUseSao = curEncoder->m_frameFilter.m_useSao = 0;
-            }
-
             if (m_param->rc.rateControlMode != X265_RC_CQP)
-                m_lookahead->getEstimatedPictureCost(frameEnc);
+                m_lookahead->getEstimatedPictureCost(frameEnc[0]);
 
             if (m_param->bIntraRefresh)
-                 calcRefreshInterval(frameEnc);
+                 calcRefreshInterval(frameEnc[0]);
 
             // Generate MCSTF References and perform HME
-            if (m_param->bEnableTemporalFilter && isFilterThisframe(frameEnc->m_mcstf->m_sliceTypeConfig, frameEnc->m_lowres.sliceType))
+            if (m_param->bEnableTemporalFilter && isFilterThisframe(frameEnc[0]->m_mcstf->m_sliceTypeConfig, frameEnc[0]->m_lowres.sliceType))
             {
 
-                if (!generateMcstfRef(frameEnc, curEncoder))
+                if (!generateMcstfRef(frameEnc[0], curEncoder))
                 {
                     m_aborted = true;
-                    x265_log(m_param, X265_LOG_ERROR, "Failed to initialize MCSTFReferencePicInfo at POC %d\n", frameEnc->m_poc);
+                    x265_log(m_param, X265_LOG_ERROR, "Failed to initialize MCSTFReferencePicInfo at POC %d\n", frameEnc[0]->m_poc);
                     fflush(stderr);
                     return -1;
                 }
 
 
-                if (!*frameEnc->m_isSubSampled)
+                if (!*frameEnc[0]->m_isSubSampled)
                 {
-                    primitives.frameSubSampleLuma((const pixel *)frameEnc->m_fencPic->m_picOrg[0],frameEnc->m_fencPicSubsampled2->m_picOrg[0], frameEnc->m_fencPic->m_stride, frameEnc->m_fencPicSubsampled2->m_stride, frameEnc->m_fencPicSubsampled2->m_picWidth, frameEnc->m_fencPicSubsampled2->m_picHeight);
-                    extendPicBorder(frameEnc->m_fencPicSubsampled2->m_picOrg[0], frameEnc->m_fencPicSubsampled2->m_stride, frameEnc->m_fencPicSubsampled2->m_picWidth, frameEnc->m_fencPicSubsampled2->m_picHeight, frameEnc->m_fencPicSubsampled2->m_lumaMarginX, frameEnc->m_fencPicSubsampled2->m_lumaMarginY);
-                    primitives.frameSubSampleLuma((const pixel *)frameEnc->m_fencPicSubsampled2->m_picOrg[0],frameEnc->m_fencPicSubsampled4->m_picOrg[0], frameEnc->m_fencPicSubsampled2->m_stride, frameEnc->m_fencPicSubsampled4->m_stride, frameEnc->m_fencPicSubsampled4->m_picWidth, frameEnc->m_fencPicSubsampled4->m_picHeight);
-                    extendPicBorder(frameEnc->m_fencPicSubsampled4->m_picOrg[0], frameEnc->m_fencPicSubsampled4->m_stride, frameEnc->m_fencPicSubsampled4->m_picWidth, frameEnc->m_fencPicSubsampled4->m_picHeight, frameEnc->m_fencPicSubsampled4->m_lumaMarginX, frameEnc->m_fencPicSubsampled4->m_lumaMarginY);
-                    *frameEnc->m_isSubSampled = true;
+                    primitives.frameSubSampleLuma((const pixel *)frameEnc[0]->m_fencPic->m_picOrg[0],frameEnc[0]->m_fencPicSubsampled2->m_picOrg[0], frameEnc[0]->m_fencPic->m_stride, frameEnc[0]->m_fencPicSubsampled2->m_stride, frameEnc[0]->m_fencPicSubsampled2->m_picWidth, frameEnc[0]->m_fencPicSubsampled2->m_picHeight);
+                    extendPicBorder(frameEnc[0]->m_fencPicSubsampled2->m_picOrg[0], frameEnc[0]->m_fencPicSubsampled2->m_stride, frameEnc[0]->m_fencPicSubsampled2->m_picWidth, frameEnc[0]->m_fencPicSubsampled2->m_picHeight, frameEnc[0]->m_fencPicSubsampled2->m_lumaMarginX, frameEnc[0]->m_fencPicSubsampled2->m_lumaMarginY);
+                    primitives.frameSubSampleLuma((const pixel *)frameEnc[0]->m_fencPicSubsampled2->m_picOrg[0],frameEnc[0]->m_fencPicSubsampled4->m_picOrg[0], frameEnc[0]->m_fencPicSubsampled2->m_stride, frameEnc[0]->m_fencPicSubsampled4->m_stride, frameEnc[0]->m_fencPicSubsampled4->m_picWidth, frameEnc[0]->m_fencPicSubsampled4->m_picHeight);
+                    extendPicBorder(frameEnc[0]->m_fencPicSubsampled4->m_picOrg[0], frameEnc[0]->m_fencPicSubsampled4->m_stride, frameEnc[0]->m_fencPicSubsampled4->m_picWidth, frameEnc[0]->m_fencPicSubsampled4->m_picHeight, frameEnc[0]->m_fencPicSubsampled4->m_lumaMarginX, frameEnc[0]->m_fencPicSubsampled4->m_lumaMarginY);
+                    *frameEnc[0]->m_isSubSampled = true;
                 }
 
-                for (uint8_t i = 1; i <= frameEnc->m_mcstf->m_numRef; i++)
+                for (uint8_t i = 1; i <= frameEnc[0]->m_mcstf->m_numRef; i++)
                 {
                     TemporalFilterRefPicInfo *ref = &curEncoder->m_mcstfRefList[i - 1];
                     if (!*ref->isSubsampled)
@@ -2408,21 +2427,21 @@ int Encoder::encode(const x265_picture* pic_in, x265_picture* pic_out)
                     }
                 }
 
-                for (uint8_t i = 1; i <= frameEnc->m_mcstf->m_numRef; i++)
+                for (uint8_t i = 1; i <= frameEnc[0]->m_mcstf->m_numRef; i++)
                 {
                     TemporalFilterRefPicInfo *ref = &curEncoder->m_mcstfRefList[i - 1];
 
-                    curEncoder->m_frameEncTF->motionEstimationLuma(ref->mvs0, ref->mvsStride0, frameEnc->m_fencPicSubsampled4, ref->picBufferSubSampled4, 16);
-                    curEncoder->m_frameEncTF->motionEstimationLuma(ref->mvs1, ref->mvsStride1, frameEnc->m_fencPicSubsampled2, ref->picBufferSubSampled2, 16, ref->mvs0, ref->mvsStride0, 2);
-                    curEncoder->m_frameEncTF->motionEstimationLuma(ref->mvs2, ref->mvsStride2, frameEnc->m_fencPic, ref->picBuffer, 16, ref->mvs1, ref->mvsStride1, 2);
-                    curEncoder->m_frameEncTF->motionEstimationLumaDoubleRes(ref->mvs,  ref->mvsStride, frameEnc->m_fencPic, ref->picBuffer, 8, ref->mvs2, ref->mvsStride2, 1, ref->error);
+                    curEncoder->m_frameEncTF->motionEstimationLuma(ref->mvs0, ref->mvsStride0, frameEnc[0]->m_fencPicSubsampled4, ref->picBufferSubSampled4, 16);
+                    curEncoder->m_frameEncTF->motionEstimationLuma(ref->mvs1, ref->mvsStride1, frameEnc[0]->m_fencPicSubsampled2, ref->picBufferSubSampled2, 16, ref->mvs0, ref->mvsStride0, 2);
+                    curEncoder->m_frameEncTF->motionEstimationLuma(ref->mvs2, ref->mvsStride2, frameEnc[0]->m_fencPic, ref->picBuffer, 16, ref->mvs1, ref->mvsStride1, 2);
+                    curEncoder->m_frameEncTF->motionEstimationLumaDoubleRes(ref->mvs,  ref->mvsStride, frameEnc[0]->m_fencPic, ref->picBuffer, 8, ref->mvs2, ref->mvsStride2, 1, ref->error);
                 }
 
-                for (int i = 0; i < frameEnc->m_mcstf->m_numRef; i++)
+                for (int i = 0; i < frameEnc[0]->m_mcstf->m_numRef; i++)
                 {
                     TemporalFilterRefPicInfo *ref = &curEncoder->m_mcstfRefList[i];
-                    ref->slicetype = m_lookahead->findSliceType(frameEnc->m_poc + ref->origOffset);
-                    Frame* dpbframePtr = m_dpb->m_picList.getPOC(frameEnc->m_poc + ref->origOffset);
+                    ref->slicetype = m_lookahead->findSliceType(frameEnc[0]->m_poc + ref->origOffset);
+                    Frame* dpbframePtr = m_dpb->m_picList.getPOC(frameEnc[0]->m_poc + ref->origOffset);
                     if (dpbframePtr != NULL)
                     {
                         if (dpbframePtr->m_encData->m_slice->m_sliceType == B_SLICE)
@@ -2557,10 +2576,10 @@ void Encoder::copyCtuInfo(x265_ctu_info_t** frameCtuInfo, int poc)
                 for (int i = 0; i < m_param->frameNumThreads; i++)
                 {
                     prevEncoder = m_frameEncoder[i];
-                    prevFrame = prevEncoder->m_frame;
-                    if (prevFrame && (prevEncoder->m_frame->m_poc == poc - 1))
+                    prevFrame = prevEncoder->m_frame[0];
+                    if (prevFrame && (prevEncoder->m_frame[0]->m_poc == poc - 1))
                     {
-                        prevFrame = prevEncoder->m_frame;
+                        prevFrame = prevEncoder->m_frame[0];
                         break;
                     }
                 }
@@ -2597,7 +2616,7 @@ void Encoder::copyCtuInfo(x265_ctu_info_t** frameCtuInfo, int poc)
             for (int i = 0; i < m_param->frameNumThreads; i++)
             {
                 curEncoder = m_frameEncoder[i];
-                curFrame = curEncoder->m_frame;
+                curFrame = curEncoder->m_frame[0];
                 if (curFrame)
                 {
                     if (poc == curFrame->m_poc)
diff --git a/source/encoder/frameencoder.cpp b/source/encoder/frameencoder.cpp
index 659b87c89..ebaf3e840 100644
--- a/source/encoder/frameencoder.cpp
+++ b/source/encoder/frameencoder.cpp
@@ -56,11 +56,12 @@ FrameEncoder::FrameEncoder()
     m_rows = NULL;
     m_top = NULL;
     m_param = NULL;
-    m_frame = NULL;
     m_cuGeoms = NULL;
     m_ctuGeomMap = NULL;
     m_localTldIdx = 0;
     memset(&m_rce, 0, sizeof(RateControlEntry));
+    for(int layer = 0; layer < MAX_SCALABLE_LAYERS; layer++)
+        m_frame[layer] = NULL;
 }
 
 void FrameEncoder::destroy()
@@ -282,14 +283,17 @@ bool FrameEncoder::initializeGeoms()
     return true;
 }
 
-bool FrameEncoder::startCompressFrame(Frame* curFrame)
+bool FrameEncoder::startCompressFrame(Frame* curFrame[MAX_SCALABLE_LAYERS])
 {
     m_slicetypeWaitTime = x265_mdate() - m_prevOutputTime;
-    m_frame = curFrame;
-    m_sliceType = curFrame->m_lowres.sliceType;
-    curFrame->m_encData->m_frameEncoderID = m_jpId;
-    curFrame->m_encData->m_jobProvider = this;
-    curFrame->m_encData->m_slice->m_mref = m_mref;
+    for (int layer = 0; layer < MAX_SCALABLE_LAYERS; layer++)
+    {
+        m_frame[layer] = curFrame[layer];
+        curFrame[layer]->m_encData->m_frameEncoderID = m_jpId;
+        curFrame[layer]->m_encData->m_jobProvider = this;
+        curFrame[layer]->m_encData->m_slice->m_mref = m_mref;
+    }
+    m_sliceType = curFrame[0]->m_lowres.sliceType;
 
     if (!m_cuGeoms)
     {
@@ -355,15 +359,17 @@ void FrameEncoder::threadMain()
     {
         if (m_param->bCTUInfo)
         {
-            while (!m_frame->m_ctuInfo)
-                m_frame->m_copied.wait();
+            while (!m_frame[0]->m_ctuInfo)
+                m_frame[0]->m_copied.wait();
         }
-        if ((m_param->bAnalysisType == AVC_INFO) && !m_param->analysisSave && !m_param->analysisLoad && !(IS_X265_TYPE_I(m_frame->m_lowres.sliceType)))
+        if ((m_param->bAnalysisType == AVC_INFO) && !m_param->analysisSave && !m_param->analysisLoad && !(IS_X265_TYPE_I(m_frame[0]->m_lowres.sliceType)))
         {
-            while (((m_frame->m_analysisData.interData == NULL && m_frame->m_analysisData.intraData == NULL) || (uint32_t)m_frame->m_poc != m_frame->m_analysisData.poc))
-                m_frame->m_copyMVType.wait();
+            while (((m_frame[0]->m_analysisData.interData == NULL && m_frame[0]->m_analysisData.intraData == NULL) || (uint32_t)m_frame[0]->m_poc != m_frame[0]->m_analysisData.poc))
+                m_frame[0]->m_copyMVType.wait();
         }
-        compressFrame();
+
+        for(int layer = 0; layer < MAX_SCALABLE_LAYERS; layer ++)
+            compressFrame(layer);
         m_done.trigger(); /* FrameEncoder::getEncodedPicture() blocks for this event */
         m_enable.wait();
     }
@@ -371,7 +377,7 @@ void FrameEncoder::threadMain()
 
 void FrameEncoder::WeightAnalysis::processTasks(int /* workerThreadId */)
 {
-    Frame* frame = master.m_frame;
+    Frame* frame = master.m_frame[master.m_sLayerId];
     weightAnalyse(*frame->m_encData->m_slice, *frame, *master.m_param);
 }
 
@@ -411,13 +417,13 @@ bool FrameEncoder::writeToneMapInfo(x265_sei_payload *payload)
         memcpy(m_top->m_prevTonemapPayload.payload, payload->payload, payload->payloadSize);
     }
 
-    bool isIDR = m_frame->m_lowres.sliceType == X265_TYPE_IDR;
+    bool isIDR = m_frame[0]->m_lowres.sliceType == X265_TYPE_IDR;
     return (payloadChange || isIDR);
 }
 
-void FrameEncoder::writeTrailingSEIMessages()
+void FrameEncoder::writeTrailingSEIMessages(int layer)
 {
-    Slice* slice = m_frame->m_encData->m_slice;
+    Slice* slice = m_frame[layer]->m_encData->m_slice;
     int planes = (m_param->internalCsp != X265_CSP_I400) ? 3 : 1;
     int32_t payloadSize = 0;
 
@@ -444,10 +450,10 @@ void FrameEncoder::writeTrailingSEIMessages()
     }
 
     m_seiReconPictureDigest.setSize(payloadSize);
-    m_seiReconPictureDigest.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_SUFFIX_SEI, m_nalList, false);
+    m_seiReconPictureDigest.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_SUFFIX_SEI, m_nalList, false, layer);
 }
 
-void FrameEncoder::compressFrame()
+void FrameEncoder::compressFrame(int layer)
 {
     ProfileScopeEvent(frameThread);
 
@@ -469,15 +475,16 @@ void FrameEncoder::compressFrame()
     m_SSDY = m_SSDU = m_SSDV = 0;
     m_ssim = 0;
     m_ssimCnt = 0;
-    memset(&(m_frame->m_encData->m_frameStats), 0, sizeof(m_frame->m_encData->m_frameStats));
+    memset(&(m_frame[layer]->m_encData->m_frameStats), 0, sizeof(m_frame[layer]->m_encData->m_frameStats));
+    m_sLayerId = layer;
 
     if (m_param->rc.aqMode != X265_AQ_EDGE && m_param->recursionSkipMode == EDGE_BASED_RSKIP)
     {
-        int height = m_frame->m_fencPic->m_picHeight;
-        int width = m_frame->m_fencPic->m_picWidth;
-        intptr_t stride = m_frame->m_fencPic->m_stride;
+        int height = m_frame[layer]->m_fencPic->m_picHeight;
+        int width = m_frame[layer]->m_fencPic->m_picWidth;
+        intptr_t stride = m_frame[layer]->m_fencPic->m_stride;
 
-        if (!computeEdge(m_frame->m_edgeBitPic, m_frame->m_fencPic->m_picOrg[0], NULL, stride, height, width, false, 1))
+        if (!computeEdge(m_frame[layer]->m_edgeBitPic, m_frame[layer]->m_fencPic->m_picOrg[0], NULL, stride, height, width, false, 1))
         {
             x265_log(m_param, X265_LOG_ERROR, " Failed to compute edge !");
         }
@@ -486,15 +493,15 @@ void FrameEncoder::compressFrame()
     /* Emit access unit delimiter unless this is the first frame and the user is
      * not repeating headers (since AUD is supposed to be the first NAL in the access
      * unit) */
-    Slice* slice = m_frame->m_encData->m_slice;
+    Slice* slice = m_frame[layer]->m_encData->m_slice;
 
-    if (m_param->bEnableEndOfSequence && m_frame->m_lowres.sliceType == X265_TYPE_IDR && m_frame->m_poc)
+    if (m_param->bEnableEndOfSequence && m_frame[layer]->m_lowres.sliceType == X265_TYPE_IDR && m_frame[layer]->m_poc)
     {
         m_bs.resetBits();
         m_nalList.serialize(NAL_UNIT_EOS, m_bs);
     }
 
-    if (m_param->bEnableAccessUnitDelimiters && (m_frame->m_poc || m_param->bRepeatHeaders))
+    if (m_param->bEnableAccessUnitDelimiters && (m_frame[layer]->m_poc || m_param->bRepeatHeaders))
     {
         m_bs.resetBits();
         m_entropyCoder.setBitstream(&m_bs);
@@ -504,7 +511,7 @@ void FrameEncoder::compressFrame()
         if (m_param->bSingleSeiNal)
             m_bs.resetBits();
     }
-    if (m_frame->m_lowres.bKeyframe && m_param->bRepeatHeaders)
+    if (m_frame[layer]->m_lowres.bKeyframe && m_param->bRepeatHeaders)
     {
         if (m_param->bOptRefListLengthPPS)
         {
@@ -526,7 +533,7 @@ void FrameEncoder::compressFrame()
     }
 
     if (m_top->m_param->rc.bStatRead && m_top->m_param->bMultiPassOptRPS)
-        m_frame->m_encData->m_slice->m_rpsIdx = (m_top->m_rateControl->m_rce2Pass + m_frame->m_encodeOrder)->rpsIdx;
+        m_frame[layer]->m_encData->m_slice->m_rpsIdx = (m_top->m_rateControl->m_rce2Pass + m_frame[layer]->m_encodeOrder)->rpsIdx;
 
     // Weighted Prediction parameters estimation.
     bool bUseWeightP = slice->m_sliceType == P_SLICE && slice->m_pps->bUseWeightPred;
@@ -534,7 +541,7 @@ void FrameEncoder::compressFrame()
 
     WeightParam* reuseWP = NULL;
     if (m_param->analysisLoad && (bUseWeightP || bUseWeightB))
-        reuseWP = (WeightParam*)m_frame->m_analysisData.wt;
+        reuseWP = (WeightParam*)m_frame[layer]->m_analysisData.wt;
 
     if (bUseWeightP || bUseWeightB)
     {
@@ -561,14 +568,14 @@ void FrameEncoder::compressFrame()
                 /* use an idle worker for weight analysis */
                 wa.waitForExit();
             else
-                weightAnalyse(*slice, *m_frame, *m_param);
+                weightAnalyse(*slice, *m_frame[layer], *m_param);
         }
     }
     else
         slice->disableWeights();
 
     if (m_param->analysisSave && (bUseWeightP || bUseWeightB))
-        reuseWP = (WeightParam*)m_frame->m_analysisData.wt;
+        reuseWP = (WeightParam*)m_frame[layer]->m_analysisData.wt;
     // Generate motion references
     int numPredDir = slice->isInterP() ? 1 : slice->isInterB() ? 2 : 0;
     for (int l = 0; l < numPredDir; l++)
@@ -597,18 +604,19 @@ void FrameEncoder::compressFrame()
 
     /* Get the QP for this frame from rate control. This call may block until
      * frames ahead of it in encode order have called rateControlEnd() */
-    int qp = m_top->m_rateControl->rateControlStart(m_frame, &m_rce, m_top);
+    int qp = (layer == 0) ? m_top->m_rateControl->rateControlStart(m_frame[layer], &m_rce, m_top) : m_rce.newQp;
+
     m_rce.newQp = qp;
 
     if (m_param->bEnableTemporalFilter)
     {
         m_frameEncTF->m_QP = qp;
-        m_frameEncTF->bilateralFilter(m_frame, m_mcstfRefList, m_param->temporalFilterStrength);
+        m_frameEncTF->bilateralFilter(m_frame[layer], m_mcstfRefList, m_param->temporalFilterStrength);
     }
 
     if (m_nr)
     {
-        if (qp > QP_MAX_SPEC && m_frame->m_param->rc.vbvBufferSize)
+        if (qp > QP_MAX_SPEC && m_frame[layer]->m_param->rc.vbvBufferSize)
         {
             for (int i = 0; i < numTLD; i++)
             {
@@ -659,11 +667,12 @@ void FrameEncoder::compressFrame()
     }
     m_initSliceContext.resetEntropy(*slice);
 
-    m_frameFilter.start(m_frame, m_initSliceContext);
+    m_frameFilter.start(m_frame[layer], m_initSliceContext);
 
     /* ensure all rows are blocked prior to initializing row CTU counters */
     WaveFront::clearEnabledRowMask();
 
+    WaveFront::setLayerId(layer);
     /* reset entropy coders and compute slice id */
     m_entropyCoder.load(m_initSliceContext);
     for (uint32_t sliceId = 0; sliceId < m_param->maxSlices; sliceId++)   
@@ -699,10 +708,10 @@ void FrameEncoder::compressFrame()
         }
     }
 
-    m_rce.encodeOrder = m_frame->m_encodeOrder;
+    m_rce.encodeOrder = m_frame[layer]->m_encodeOrder;
     int prevBPSEI = m_rce.encodeOrder ? m_top->m_lastBPSEI : 0;
 
-    if (m_frame->m_lowres.bKeyframe)
+    if (m_frame[layer]->m_lowres.bKeyframe)
     {
         if (m_param->bEmitHRDSEI)
         {
@@ -712,23 +721,23 @@ void FrameEncoder::compressFrame()
             bpSei->m_auCpbRemovalDelayDelta = 1;
             bpSei->m_cpbDelayOffset = 0;
             bpSei->m_dpbDelayOffset = 0;
-            bpSei->m_concatenationFlag = (m_param->bEnableHRDConcatFlag && !m_frame->m_poc) ? true : false;
+            bpSei->m_concatenationFlag = (m_param->bEnableHRDConcatFlag && !m_frame[layer]->m_poc) ? true : false;
 
             // hrdFullness() calculates the initial CPB removal delay and offset
             m_top->m_rateControl->hrdFullness(bpSei);
-            bpSei->writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal);
+            bpSei->writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal, layer);
 
             m_top->m_lastBPSEI = m_rce.encodeOrder;
         }
 
-        if (m_frame->m_lowres.sliceType == X265_TYPE_IDR && m_param->bEmitIDRRecoverySEI)
+        if (m_frame[layer]->m_lowres.sliceType == X265_TYPE_IDR && m_param->bEmitIDRRecoverySEI)
         {
             /* Recovery Point SEI require the SPS to be "activated" */
             SEIRecoveryPoint sei;
             sei.m_recoveryPocCnt = 0;
             sei.m_exactMatchingFlag = true;
             sei.m_brokenLinkFlag = false;
-            sei.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal);
+            sei.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal, layer);
         }
     }
 
@@ -761,7 +770,7 @@ void FrameEncoder::compressFrame()
                 }
             }
             else if (m_param->bEnableFrameDuplication)
-                sei->m_picStruct = m_frame->m_picStruct;
+                sei->m_picStruct = m_frame[layer]->m_picStruct;
             else
                 sei->m_picStruct = m_param->pictureStructure;
 
@@ -777,17 +786,17 @@ void FrameEncoder::compressFrame()
             // wait after removal of the access unit with the most recent
             // buffering period SEI message
             sei->m_auCpbRemovalDelay = X265_MIN(X265_MAX(1, m_rce.encodeOrder - prevBPSEI), (1 << hrd->cpbRemovalDelayLength));
-            sei->m_picDpbOutputDelay = slice->m_sps->numReorderPics[m_frame->m_tempLayer] + poc - m_rce.encodeOrder;
+            sei->m_picDpbOutputDelay = slice->m_sps->numReorderPics[m_frame[layer]->m_tempLayer] + poc - m_rce.encodeOrder;
         }
 
-        sei->writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal);
+        sei->writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal, layer);
     }
 
     if (m_param->preferredTransferCharacteristics > -1 && slice->isIRAP())
     {
         SEIAlternativeTC m_seiAlternativeTC;
         m_seiAlternativeTC.m_preferredTransferCharacteristics = m_param->preferredTransferCharacteristics;
-        m_seiAlternativeTC.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal);
+        m_seiAlternativeTC.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal, layer);
     }
     /* Write Film grain characteristics if present */
     if (this->m_top->m_filmGrainIn)
@@ -795,18 +804,18 @@ void FrameEncoder::compressFrame()
         FilmGrainCharacteristics m_filmGrain;
         /* Read the Film grain model file */
         readModel(&m_filmGrain, this->m_top->m_filmGrainIn);
-        m_filmGrain.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal);
+        m_filmGrain.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal, layer);
     }
     /* Write user SEI */
-    for (int i = 0; i < m_frame->m_userSEI.numPayloads; i++)
+    for (int i = 0; i < m_frame[layer]->m_userSEI.numPayloads; i++)
     {
-        x265_sei_payload *payload = &m_frame->m_userSEI.payloads[i];
+        x265_sei_payload *payload = &m_frame[layer]->m_userSEI.payloads[i];
         if (payload->payloadType == USER_DATA_UNREGISTERED)
         {
             SEIuserDataUnregistered sei;
             sei.m_userData = payload->payload;
             sei.setSize(payload->payloadSize);
-            sei.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal);
+            sei.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal, layer);
         }
         else if (payload->payloadType == USER_DATA_REGISTERED_ITU_T_T35)
         {
@@ -816,16 +825,16 @@ void FrameEncoder::compressFrame()
                 SEIuserDataRegistered sei;
                 sei.m_userData = payload->payload;
                 sei.setSize(payload->payloadSize);
-                sei.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal);
+                sei.writeSEImessages(m_bs, *slice->m_sps, NAL_UNIT_PREFIX_SEI, m_nalList, m_param->bSingleSeiNal, layer);
             }
         }
         else
             x265_log(m_param, X265_LOG_ERROR, "Unrecognized SEI type\n");
     }
 
-    bool isSei = ((m_frame->m_lowres.bKeyframe && m_param->bRepeatHeaders) || m_param->bEmitHRDSEI ||
-                 !!m_param->interlaceMode || (m_frame->m_lowres.sliceType == X265_TYPE_IDR && m_param->bEmitIDRRecoverySEI) ||
-                   m_frame->m_userSEI.numPayloads);
+    bool isSei = ((m_frame[layer]->m_lowres.bKeyframe && m_param->bRepeatHeaders) || m_param->bEmitHRDSEI ||
+                 !!m_param->interlaceMode || (m_frame[layer]->m_lowres.sliceType == X265_TYPE_IDR && m_param->bEmitIDRRecoverySEI) ||
+                   m_frame[layer]->m_userSEI.numPayloads);
 
     if (isSei && m_param->bSingleSeiNal)
     {
@@ -846,7 +855,7 @@ void FrameEncoder::compressFrame()
     }
 
     if (m_param->bDynamicRefine)
-        computeAvgTrainingData();
+        computeAvgTrainingData(layer);
 
     /* Analyze CTU rows, most of the hard work is done here.  Frame is
      * compressed in a wave-front pattern if WPP is enabled. Row based loop
@@ -951,12 +960,12 @@ void FrameEncoder::compressFrame()
                     m_row0WaitTime = x265_mdate();
                 else if (i == m_numRows - 1)
                     m_allRowsAvailableTime = x265_mdate();
-                processRowEncoder(i, m_tld[m_localTldIdx]);
+                processRowEncoder(i, m_tld[m_localTldIdx], layer);
             }
 
             // filter
             if (i >= m_filterRowDelay)
-                m_frameFilter.processRow(i - m_filterRowDelay);
+                m_frameFilter.processRow(i - m_filterRowDelay, layer);
         }
     }
 #if ENABLE_LIBVMAF
@@ -965,15 +974,15 @@ void FrameEncoder::compressFrame()
 
     if (m_param->maxSlices > 1)
     {
-        PicYuv *reconPic = m_frame->m_reconPic;
+        PicYuv *reconPic = m_frame[layer]->m_reconPic;
         uint32_t height = reconPic->m_picHeight;
-        initDecodedPictureHashSEI(0, 0, height);
+        initDecodedPictureHashSEI(0, 0, height, layer);
     } 
 
-    if (m_param->bDynamicRefine && m_top->m_startPoint <= m_frame->m_encodeOrder) //Avoid collecting data that will not be used by future frames.
-        collectDynDataFrame();
+    if (m_param->bDynamicRefine && m_top->m_startPoint <= m_frame[layer]->m_encodeOrder) //Avoid collecting data that will not be used by future frames.
+        collectDynDataFrame(layer);
 
-    if (m_param->bEnableTemporalFilter && m_top->isFilterThisframe(m_frame->m_mcstf->m_sliceTypeConfig, m_frame->m_lowres.sliceType))
+    if (m_param->bEnableTemporalFilter && m_top->isFilterThisframe(m_frame[layer]->m_mcstf->m_sliceTypeConfig, m_frame[layer]->m_lowres.sliceType))
     {
         //Reset the MCSTF context in Frame Encoder and Frame
         for (int i = 0; i < (m_frameEncTF->m_range << 1); i++)
@@ -985,7 +994,7 @@ void FrameEncoder::compressFrame()
             memset(m_mcstfRefList[i].noise, 0, sizeof(int) * ((m_param->sourceWidth / 4) * (m_param->sourceHeight / 4)));
             memset(m_mcstfRefList[i].error, 0, sizeof(int) * ((m_param->sourceWidth / 4) * (m_param->sourceHeight / 4)));
 
-            m_frame->m_mcstf->m_numRef = 0;
+            m_frame[layer]->m_mcstf->m_numRef = 0;
         }
     }
 
@@ -997,64 +1006,64 @@ void FrameEncoder::compressFrame()
         // accumulate intra,inter,skip cu count per frame for 2 pass
         for (uint32_t i = 0; i < m_numRows; i++)
         {
-            m_frame->m_encData->m_frameStats.mvBits    += m_rows[i].rowStats.mvBits;
-            m_frame->m_encData->m_frameStats.coeffBits += m_rows[i].rowStats.coeffBits;
-            m_frame->m_encData->m_frameStats.miscBits  += m_rows[i].rowStats.miscBits;
+            m_frame[layer]->m_encData->m_frameStats.mvBits    += m_rows[i].rowStats.mvBits;
+            m_frame[layer]->m_encData->m_frameStats.coeffBits += m_rows[i].rowStats.coeffBits;
+            m_frame[layer]->m_encData->m_frameStats.miscBits  += m_rows[i].rowStats.miscBits;
             totalI                                     += m_rows[i].rowStats.intra8x8Cnt;
             totalP                                     += m_rows[i].rowStats.inter8x8Cnt;
             totalSkip                                  += m_rows[i].rowStats.skip8x8Cnt;
         }
         int totalCuCount = totalI + totalP + totalSkip;
-        m_frame->m_encData->m_frameStats.percent8x8Intra = (double)totalI / totalCuCount;
-        m_frame->m_encData->m_frameStats.percent8x8Inter = (double)totalP / totalCuCount;
-        m_frame->m_encData->m_frameStats.percent8x8Skip  = (double)totalSkip / totalCuCount;
+        m_frame[layer]->m_encData->m_frameStats.percent8x8Intra = (double)totalI / totalCuCount;
+        m_frame[layer]->m_encData->m_frameStats.percent8x8Inter = (double)totalP / totalCuCount;
+        m_frame[layer]->m_encData->m_frameStats.percent8x8Skip  = (double)totalSkip / totalCuCount;
     }
 
     if (m_param->csvLogLevel >= 1)
     {
         for (uint32_t i = 0; i < m_numRows; i++)
         {
-            m_frame->m_encData->m_frameStats.cntIntraNxN += m_rows[i].rowStats.cntIntraNxN;
-            m_frame->m_encData->m_frameStats.totalCu += m_rows[i].rowStats.totalCu;
-            m_frame->m_encData->m_frameStats.totalCtu += m_rows[i].rowStats.totalCtu;
-            m_frame->m_encData->m_frameStats.lumaDistortion += m_rows[i].rowStats.lumaDistortion;
-            m_frame->m_encData->m_frameStats.chromaDistortion += m_rows[i].rowStats.chromaDistortion;
-            m_frame->m_encData->m_frameStats.psyEnergy += m_rows[i].rowStats.psyEnergy;
-            m_frame->m_encData->m_frameStats.ssimEnergy += m_rows[i].rowStats.ssimEnergy;
-            m_frame->m_encData->m_frameStats.resEnergy += m_rows[i].rowStats.resEnergy;
+            m_frame[layer]->m_encData->m_frameStats.cntIntraNxN += m_rows[i].rowStats.cntIntraNxN;
+            m_frame[layer]->m_encData->m_frameStats.totalCu += m_rows[i].rowStats.totalCu;
+            m_frame[layer]->m_encData->m_frameStats.totalCtu += m_rows[i].rowStats.totalCtu;
+            m_frame[layer]->m_encData->m_frameStats.lumaDistortion += m_rows[i].rowStats.lumaDistortion;
+            m_frame[layer]->m_encData->m_frameStats.chromaDistortion += m_rows[i].rowStats.chromaDistortion;
+            m_frame[layer]->m_encData->m_frameStats.psyEnergy += m_rows[i].rowStats.psyEnergy;
+            m_frame[layer]->m_encData->m_frameStats.ssimEnergy += m_rows[i].rowStats.ssimEnergy;
+            m_frame[layer]->m_encData->m_frameStats.resEnergy += m_rows[i].rowStats.resEnergy;
             for (uint32_t depth = 0; depth <= m_param->maxCUDepth; depth++)
             {
-                m_frame->m_encData->m_frameStats.cntSkipCu[depth] += m_rows[i].rowStats.cntSkipCu[depth];
-                m_frame->m_encData->m_frameStats.cntMergeCu[depth] += m_rows[i].rowStats.cntMergeCu[depth];
+                m_frame[layer]->m_encData->m_frameStats.cntSkipCu[depth] += m_rows[i].rowStats.cntSkipCu[depth];
+                m_frame[layer]->m_encData->m_frameStats.cntMergeCu[depth] += m_rows[i].rowStats.cntMergeCu[depth];
                 for (int m = 0; m < INTER_MODES; m++)
-                    m_frame->m_encData->m_frameStats.cuInterDistribution[depth][m] += m_rows[i].rowStats.cuInterDistribution[depth][m];
+                    m_frame[layer]->m_encData->m_frameStats.cuInterDistribution[depth][m] += m_rows[i].rowStats.cuInterDistribution[depth][m];
                 for (int n = 0; n < INTRA_MODES; n++)
-                    m_frame->m_encData->m_frameStats.cuIntraDistribution[depth][n] += m_rows[i].rowStats.cuIntraDistribution[depth][n];
+                    m_frame[layer]->m_encData->m_frameStats.cuIntraDistribution[depth][n] += m_rows[i].rowStats.cuIntraDistribution[depth][n];
             }
         }
-        m_frame->m_encData->m_frameStats.percentIntraNxN = (double)(m_frame->m_encData->m_frameStats.cntIntraNxN * 100) / m_frame->m_encData->m_frameStats.totalCu;
+        m_frame[layer]->m_encData->m_frameStats.percentIntraNxN = (double)(m_frame[layer]->m_encData->m_frameStats.cntIntraNxN * 100) / m_frame[layer]->m_encData->m_frameStats.totalCu;
 
         for (uint32_t depth = 0; depth <= m_param->maxCUDepth; depth++)
         {
-            m_frame->m_encData->m_frameStats.percentSkipCu[depth] = (double)(m_frame->m_encData->m_frameStats.cntSkipCu[depth] * 100) / m_frame->m_encData->m_frameStats.totalCu;
-            m_frame->m_encData->m_frameStats.percentMergeCu[depth] = (double)(m_frame->m_encData->m_frameStats.cntMergeCu[depth] * 100) / m_frame->m_encData->m_frameStats.totalCu;
+            m_frame[layer]->m_encData->m_frameStats.percentSkipCu[depth] = (double)(m_frame[layer]->m_encData->m_frameStats.cntSkipCu[depth] * 100) / m_frame[layer]->m_encData->m_frameStats.totalCu;
+            m_frame[layer]->m_encData->m_frameStats.percentMergeCu[depth] = (double)(m_frame[layer]->m_encData->m_frameStats.cntMergeCu[depth] * 100) / m_frame[layer]->m_encData->m_frameStats.totalCu;
             for (int n = 0; n < INTRA_MODES; n++)
-                m_frame->m_encData->m_frameStats.percentIntraDistribution[depth][n] = (double)(m_frame->m_encData->m_frameStats.cuIntraDistribution[depth][n] * 100) / m_frame->m_encData->m_frameStats.totalCu;
+                m_frame[layer]->m_encData->m_frameStats.percentIntraDistribution[depth][n] = (double)(m_frame[layer]->m_encData->m_frameStats.cuIntraDistribution[depth][n] * 100) / m_frame[layer]->m_encData->m_frameStats.totalCu;
             uint64_t cuInterRectCnt = 0; // sum of Nx2N, 2NxN counts
-            cuInterRectCnt += m_frame->m_encData->m_frameStats.cuInterDistribution[depth][1] + m_frame->m_encData->m_frameStats.cuInterDistribution[depth][2];
-            m_frame->m_encData->m_frameStats.percentInterDistribution[depth][0] = (double)(m_frame->m_encData->m_frameStats.cuInterDistribution[depth][0] * 100) / m_frame->m_encData->m_frameStats.totalCu;
-            m_frame->m_encData->m_frameStats.percentInterDistribution[depth][1] = (double)(cuInterRectCnt * 100) / m_frame->m_encData->m_frameStats.totalCu;
-            m_frame->m_encData->m_frameStats.percentInterDistribution[depth][2] = (double)(m_frame->m_encData->m_frameStats.cuInterDistribution[depth][3] * 100) / m_frame->m_encData->m_frameStats.totalCu;
+            cuInterRectCnt += m_frame[layer]->m_encData->m_frameStats.cuInterDistribution[depth][1] + m_frame[layer]->m_encData->m_frameStats.cuInterDistribution[depth][2];
+            m_frame[layer]->m_encData->m_frameStats.percentInterDistribution[depth][0] = (double)(m_frame[layer]->m_encData->m_frameStats.cuInterDistribution[depth][0] * 100) / m_frame[layer]->m_encData->m_frameStats.totalCu;
+            m_frame[layer]->m_encData->m_frameStats.percentInterDistribution[depth][1] = (double)(cuInterRectCnt * 100) / m_frame[layer]->m_encData->m_frameStats.totalCu;
+            m_frame[layer]->m_encData->m_frameStats.percentInterDistribution[depth][2] = (double)(m_frame[layer]->m_encData->m_frameStats.cuInterDistribution[depth][3] * 100) / m_frame[layer]->m_encData->m_frameStats.totalCu;
         }
     }
 
     if (m_param->csvLogLevel >= 2)
     {
-        m_frame->m_encData->m_frameStats.avgLumaDistortion = (double)(m_frame->m_encData->m_frameStats.lumaDistortion) / m_frame->m_encData->m_frameStats.totalCtu;
-        m_frame->m_encData->m_frameStats.avgChromaDistortion = (double)(m_frame->m_encData->m_frameStats.chromaDistortion) / m_frame->m_encData->m_frameStats.totalCtu;
-        m_frame->m_encData->m_frameStats.avgPsyEnergy = (double)(m_frame->m_encData->m_frameStats.psyEnergy) / m_frame->m_encData->m_frameStats.totalCtu;
-        m_frame->m_encData->m_frameStats.avgSsimEnergy = (double)(m_frame->m_encData->m_frameStats.ssimEnergy) / m_frame->m_encData->m_frameStats.totalCtu;
-        m_frame->m_encData->m_frameStats.avgResEnergy = (double)(m_frame->m_encData->m_frameStats.resEnergy) / m_frame->m_encData->m_frameStats.totalCtu;
+        m_frame[layer]->m_encData->m_frameStats.avgLumaDistortion = (double)(m_frame[layer]->m_encData->m_frameStats.lumaDistortion) / m_frame[layer]->m_encData->m_frameStats.totalCtu;
+        m_frame[layer]->m_encData->m_frameStats.avgChromaDistortion = (double)(m_frame[layer]->m_encData->m_frameStats.chromaDistortion) / m_frame[layer]->m_encData->m_frameStats.totalCtu;
+        m_frame[layer]->m_encData->m_frameStats.avgPsyEnergy = (double)(m_frame[layer]->m_encData->m_frameStats.psyEnergy) / m_frame[layer]->m_encData->m_frameStats.totalCtu;
+        m_frame[layer]->m_encData->m_frameStats.avgSsimEnergy = (double)(m_frame[layer]->m_encData->m_frameStats.ssimEnergy) / m_frame[layer]->m_encData->m_frameStats.totalCtu;
+        m_frame[layer]->m_encData->m_frameStats.avgResEnergy = (double)(m_frame[layer]->m_encData->m_frameStats.resEnergy) / m_frame[layer]->m_encData->m_frameStats.totalCtu;
     }
 
     m_bs.resetBits();
@@ -1063,7 +1072,7 @@ void FrameEncoder::compressFrame()
 
     // finish encode of each CTU row, only required when SAO is enabled
     if (slice->m_bUseSao)
-        encodeSlice(0);
+        encodeSlice(0, layer);
 
     m_entropyCoder.setBitstream(&m_bs);
 
@@ -1081,7 +1090,7 @@ void FrameEncoder::compressFrame()
                 ScopedLock refIdxLock(m_top->m_sliceRefIdxLock);
                 m_top->analyseRefIdx(slice->m_numRefIdx);
             }
-            m_entropyCoder.codeSliceHeader(*slice, *m_frame->m_encData, sliceAddr, m_sliceAddrBits, slice->m_sliceQp);
+            m_entropyCoder.codeSliceHeader(*slice, *m_frame[layer]->m_encData, sliceAddr, m_sliceAddrBits, slice->m_sliceQp);
 
             // Find rows of current slice
             const uint32_t prevSliceRow = nextSliceRow;
@@ -1098,7 +1107,7 @@ void FrameEncoder::compressFrame()
             
             m_bs.writeByteAlignment();
 
-            m_nalList.serialize(slice->m_nalUnitType, m_bs, (!!m_param->bEnableTemporalSubLayers ? m_frame->m_tempLayer + 1 : (1 + (slice->m_nalUnitType == NAL_UNIT_CODED_SLICE_TSA_N))));
+            m_nalList.serialize(slice->m_nalUnitType, m_bs, layer, (!!m_param->bEnableTemporalSubLayers ? m_frame[layer]->m_tempLayer + 1 : (1 + (slice->m_nalUnitType == NAL_UNIT_CODED_SLICE_TSA_N))));
         }
     }
     else
@@ -1108,7 +1117,7 @@ void FrameEncoder::compressFrame()
             ScopedLock refIdxLock(m_top->m_sliceRefIdxLock);
             m_top->analyseRefIdx(slice->m_numRefIdx);
         }
-        m_entropyCoder.codeSliceHeader(*slice, *m_frame->m_encData, 0, 0, slice->m_sliceQp);
+        m_entropyCoder.codeSliceHeader(*slice, *m_frame[layer]->m_encData, 0, 0, slice->m_sliceQp);
 
         // serialize each row, record final lengths in slice header
         uint32_t maxStreamSize = m_nalList.serializeSubstreams(m_substreamSizes, numSubstreams, m_outStreams);
@@ -1119,11 +1128,11 @@ void FrameEncoder::compressFrame()
             m_entropyCoder.codeSliceHeaderWPPEntryPoints(m_substreamSizes, (slice->m_sps->numCuInHeight - 1), maxStreamSize);
         m_bs.writeByteAlignment();
 
-        m_nalList.serialize(slice->m_nalUnitType, m_bs, (!!m_param->bEnableTemporalSubLayers ? m_frame->m_tempLayer + 1 : (1 + (slice->m_nalUnitType == NAL_UNIT_CODED_SLICE_TSA_N))));
+        m_nalList.serialize(slice->m_nalUnitType, m_bs, layer, (!!m_param->bEnableTemporalSubLayers ? m_frame[layer]->m_tempLayer + 1 : (1 + (slice->m_nalUnitType == NAL_UNIT_CODED_SLICE_TSA_N))));
     }
 
     if (m_param->decodedPictureHashSEI)
-        writeTrailingSEIMessages();
+        writeTrailingSEIMessages(layer);
 
     uint64_t bytes = 0;
     for (uint32_t i = 0; i < m_nalList.m_numNal; i++)
@@ -1142,7 +1151,7 @@ void FrameEncoder::compressFrame()
 
     int filler = 0;
     /* rateControlEnd may also block for earlier frames to call rateControlUpdateStats */
-    if (m_top->m_rateControl->rateControlEnd(m_frame, m_accessUnitBits, &m_rce, &filler) < 0)
+    if (!layer && m_top->m_rateControl->rateControlEnd(m_frame[layer], m_accessUnitBits, &m_rce, &filler) < 0)
         m_top->m_aborted = true;
 
     if (filler > 0)
@@ -1161,11 +1170,11 @@ void FrameEncoder::compressFrame()
         m_accessUnitBits = bytes << 3;
     }
 
-    if (m_frame->m_rpu.payloadSize)
+    if (m_frame[layer]->m_rpu.payloadSize)
     {
         m_bs.resetBits();
-        for (int i = 0; i < m_frame->m_rpu.payloadSize; i++)
-            m_bs.write(m_frame->m_rpu.payload[i], 8);
+        for (int i = 0; i < m_frame[layer]->m_rpu.payloadSize; i++)
+            m_bs.write(m_frame[layer]->m_rpu.payload[i], 8);
         m_nalList.serialize(NAL_UNIT_UNSPECIFIED, m_bs);
     }
 
@@ -1223,9 +1232,9 @@ void FrameEncoder::compressFrame()
     m_endFrameTime = x265_mdate();  
 }
 
-void FrameEncoder::initDecodedPictureHashSEI(int row, int cuAddr, int height)
+void FrameEncoder::initDecodedPictureHashSEI(int row, int cuAddr, int height, int layer)
 {
-    PicYuv *reconPic = m_frame->m_reconPic;
+    PicYuv *reconPic = m_frame[layer]->m_reconPic;
     uint32_t width = reconPic->m_picWidth;	
     intptr_t stride = reconPic->m_stride;
     uint32_t maxCUHeight = m_param->maxCUSize;
@@ -1295,20 +1304,20 @@ void FrameEncoder::initDecodedPictureHashSEI(int row, int cuAddr, int height)
     }
 }
 
-void FrameEncoder::encodeSlice(uint32_t sliceAddr)
+void FrameEncoder::encodeSlice(uint32_t sliceAddr, int layer)
 {
-    Slice* slice = m_frame->m_encData->m_slice;
+    Slice* slice = m_frame[layer]->m_encData->m_slice;
     const uint32_t widthInLCUs = slice->m_sps->numCuInWidth;
     const uint32_t lastCUAddr = (slice->m_endCUAddr + m_param->num4x4Partitions - 1) / m_param->num4x4Partitions;
     const uint32_t numSubstreams = m_param->bEnableWavefront ? slice->m_sps->numCuInHeight : 1;
 
-    SAOParam* saoParam = slice->m_sps->bUseSAO && slice->m_bUseSao ? m_frame->m_encData->m_saoParam : NULL;
+    SAOParam* saoParam = slice->m_sps->bUseSAO && slice->m_bUseSao ? m_frame[layer]->m_encData->m_saoParam : NULL;
     for (uint32_t cuAddr = sliceAddr; cuAddr < lastCUAddr; cuAddr++)
     {
         uint32_t col = cuAddr % widthInLCUs;
         uint32_t row = cuAddr / widthInLCUs;
         uint32_t subStrm = row % numSubstreams;
-        CUData* ctu = m_frame->m_encData->getPicCTU(cuAddr);
+        CUData* ctu = m_frame[layer]->m_encData->getPicCTU(cuAddr);
 
         m_entropyCoder.setBitstream(&m_outStreams[subStrm]);
 
@@ -1369,7 +1378,7 @@ void FrameEncoder::encodeSlice(uint32_t sliceAddr)
         m_entropyCoder.finishSlice();
 }
 
-void FrameEncoder::processRow(int row, int threadId)
+void FrameEncoder::processRow(int row, int threadId, int layer)
 {
     int64_t startTime = x265_mdate();
     if (ATOMIC_INC(&m_activeWorkerCount) == 1 && m_stallStartTime)
@@ -1379,10 +1388,10 @@ void FrameEncoder::processRow(int row, int threadId)
     const uint32_t typeNum = m_idx_to_row[row & 1];
 
     if (!typeNum)
-        processRowEncoder(realRow, m_tld[threadId]);
+        processRowEncoder(realRow, m_tld[threadId], layer);
     else
     {
-        m_frameFilter.processRow(realRow);
+        m_frameFilter.processRow(realRow, layer);
 
         // NOTE: Active next row
         if (realRow != m_sliceBaseRow[m_rows[realRow].sliceId + 1] - 1)
@@ -1396,7 +1405,7 @@ void FrameEncoder::processRow(int row, int threadId)
 }
 
 // Called by worker threads
-void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
+void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld, int layer)
 {
     const uint32_t row = (uint32_t)intRow;
     CTURow& curRow = m_rows[row];
@@ -1424,7 +1433,7 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
     /* When WPP is enabled, every row has its own row coder instance. Otherwise
      * they share row 0 */
     Entropy& rowCoder = m_param->bEnableWavefront ? curRow.rowGoOnCoder : m_rows[0].rowGoOnCoder;
-    FrameData& curEncData = *m_frame->m_encData;
+    FrameData& curEncData = *m_frame[layer]->m_encData;
     Slice *slice = curEncData.m_slice;
 
     const uint32_t numCols = m_numCols;
@@ -1432,7 +1441,7 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
     bool bIsVbv = m_param->rc.vbvBufferSize > 0 && m_param->rc.vbvMaxBitrate > 0;
 
     const uint32_t sliceId = curRow.sliceId;
-    uint32_t maxBlockCols = (m_frame->m_fencPic->m_picWidth + (16 - 1)) / 16;
+    uint32_t maxBlockCols = (m_frame[layer]->m_fencPic->m_picWidth + (16 - 1)) / 16;
     uint32_t noOfBlocks = m_param->maxCUSize / 16;
     const uint32_t bFirstRowInSlice = ((row == 0) || (m_rows[row - 1].sliceId != curRow.sliceId)) ? 1 : 0;
     const uint32_t bLastRowInSlice = ((row == m_numRows - 1) || (m_rows[row + 1].sliceId != curRow.sliceId)) ? 1 : 0;
@@ -1452,30 +1461,30 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
             if (m_param->bEnableWavefront || !row)
             {
                 double meanQPOff = 0;
-                bool isReferenced = IS_REFERENCED(m_frame);
-                double *qpoffs = (isReferenced && m_param->rc.cuTree) ? m_frame->m_lowres.qpCuTreeOffset : m_frame->m_lowres.qpAqOffset;
+                bool isReferenced = IS_REFERENCED(m_frame[layer]);
+                double *qpoffs = (isReferenced && m_param->rc.cuTree) ? m_frame[layer]->m_lowres.qpCuTreeOffset : m_frame[layer]->m_lowres.qpAqOffset;
                 if (qpoffs)
                 {
                     uint32_t loopIncr = (m_param->rc.qgSize == 8) ? 8 : 16;
 
-                    uint32_t cuYStart = 0, height = m_frame->m_fencPic->m_picHeight;
+                    uint32_t cuYStart = 0, height = m_frame[layer]->m_fencPic->m_picHeight;
                     if (m_param->bEnableWavefront)
                     {
                         cuYStart = intRow * m_param->maxCUSize;
                         height = cuYStart + m_param->maxCUSize;
                     }
 
-                    uint32_t qgSize = m_param->rc.qgSize, width = m_frame->m_fencPic->m_picWidth;
-                    uint32_t maxOffsetCols = (m_frame->m_fencPic->m_picWidth + (loopIncr - 1)) / loopIncr;
+                    uint32_t qgSize = m_param->rc.qgSize, width = m_frame[layer]->m_fencPic->m_picWidth;
+                    uint32_t maxOffsetCols = (m_frame[layer]->m_fencPic->m_picWidth + (loopIncr - 1)) / loopIncr;
                     uint32_t count = 0;
-                    for (uint32_t cuY = cuYStart; cuY < height && (cuY < m_frame->m_fencPic->m_picHeight); cuY += qgSize)
+                    for (uint32_t cuY = cuYStart; cuY < height && (cuY < m_frame[layer]->m_fencPic->m_picHeight); cuY += qgSize)
                     {
                         for (uint32_t cuX = 0; cuX < width; cuX += qgSize)
                         {
                             double qp_offset = 0;
                             uint32_t cnt = 0;
 
-                            for (uint32_t block_yy = cuY; block_yy < cuY + qgSize && block_yy < m_frame->m_fencPic->m_picHeight; block_yy += loopIncr)
+                            for (uint32_t block_yy = cuY; block_yy < cuY + qgSize && block_yy < m_frame[layer]->m_fencPic->m_picHeight; block_yy += loopIncr)
                             {
                                 for (uint32_t block_xx = cuX; block_xx < cuX + qgSize && block_xx < width; block_xx += loopIncr)
                                 {
@@ -1518,7 +1527,7 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
         const uint32_t cuAddr = lineStartCUAddr + col;
         CUData* ctu = curEncData.getPicCTU(cuAddr);
         const uint32_t bLastCuInSlice = (bLastRowInSlice & (col == numCols - 1)) ? 1 : 0;
-        ctu->initCTU(*m_frame, cuAddr, slice->m_sliceQp, bFirstRowInSlice, bLastRowInSlice, bLastCuInSlice);
+        ctu->initCTU(*m_frame[layer], cuAddr, slice->m_sliceQp, bFirstRowInSlice, bLastRowInSlice, bLastCuInSlice);
 
         if (bIsVbv)
         {
@@ -1556,8 +1565,8 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
 
                     for (uint32_t w = 0; w < noOfBlocks && (block_x + w) < maxBlockCols; w++, idx++)
                     {
-                        cuStat.vbvCost += m_frame->m_lowres.lowresCostForRc[idx] & LOWRES_COST_MASK;
-                        cuStat.intraVbvCost += m_frame->m_lowres.intraCost[idx];
+                        cuStat.vbvCost += m_frame[layer]->m_lowres.lowresCostForRc[idx] & LOWRES_COST_MASK;
+                        cuStat.intraVbvCost += m_frame[layer]->m_lowres.intraCost[idx];
                     }
                 }
             }
@@ -1575,12 +1584,12 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
             ctu->m_vbvAffected = true;
 
         // Does all the CU analysis, returns best top level mode decision
-        Mode& best = tld.analysis.compressCTU(*ctu, *m_frame, m_cuGeoms[m_ctuGeomMap[cuAddr]], rowCoder);
+        Mode& best = tld.analysis.compressCTU(*ctu, *m_frame[layer], m_cuGeoms[m_ctuGeomMap[cuAddr]], rowCoder);
 
         /* startPoint > encodeOrder is true when the start point changes for
         a new GOP but few frames from the previous GOP is still incomplete.
         The data of frames in this interval will not be used by any future frames. */
-        if (m_param->bDynamicRefine && m_top->m_startPoint <= m_frame->m_encodeOrder)
+        if (m_param->bDynamicRefine && m_top->m_startPoint <= m_frame[layer]->m_encodeOrder)
             collectDynDataRow(*ctu, &curRow.rowStats);
 
         // take a sample of the current active worker count
@@ -1597,7 +1606,7 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
 
         /* SAO parameter estimation using non-deblocked pixels for CTU bottom and right boundary areas */
         if (slice->m_bUseSao && m_param->bSaoNonDeblocked)
-            m_frameFilter.m_parallelFilter[row].m_sao.calcSaoStatsCu_BeforeDblk(m_frame, col, row);
+            m_frameFilter.m_parallelFilter[row].m_sao.calcSaoStatsCu_BeforeDblk(m_frame[layer], col, row);
 
         /* Deblock with idle threading */
         if (m_param->bEnableLoopFilter | slice->m_bUseSao)
@@ -1705,14 +1714,14 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
             if (!m_param->bEnableWavefront && col == numCols - 1)
             {
                 double qpBase = curEncData.m_cuStat[cuAddr].baseQp;
-                curRow.reEncode = m_top->m_rateControl->rowVbvRateControl(m_frame, row, &m_rce, qpBase, m_sliceBaseRow, sliceId);
+                curRow.reEncode = m_top->m_rateControl->rowVbvRateControl(m_frame[layer], row, &m_rce, qpBase, m_sliceBaseRow, sliceId);
                 qpBase = x265_clip3((double)m_param->rc.qpMin, (double)m_param->rc.qpMax, qpBase);
                 curEncData.m_rowStat[row].rowQp = qpBase;
                 curEncData.m_rowStat[row].rowQpScale = x265_qp2qScale(qpBase);
                 if (curRow.reEncode < 0)
                 {
                     x265_log(m_param, X265_LOG_DEBUG, "POC %d row %d - encode restart required for VBV, to %.2f from %.2f\n",
-                        m_frame->m_poc, row, qpBase, curEncData.m_cuStat[cuAddr].baseQp);
+                        m_frame[layer]->m_poc, row, qpBase, curEncData.m_cuStat[cuAddr].baseQp);
 
                     m_vbvResetTriggerRow[curRow.sliceId] = row;
                     m_outStreams[0].copyBits(&m_backupStreams[0]);
@@ -1755,7 +1764,7 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
                     }
                 }
                 double qpBase = curEncData.m_cuStat[cuAddr].baseQp;
-                curRow.reEncode = m_top->m_rateControl->rowVbvRateControl(m_frame, row, &m_rce, qpBase, m_sliceBaseRow, sliceId);
+                curRow.reEncode = m_top->m_rateControl->rowVbvRateControl(m_frame[layer], row, &m_rce, qpBase, m_sliceBaseRow, sliceId);
                 qpBase = x265_clip3((double)m_param->rc.qpMin, (double)m_param->rc.qpMax, qpBase);
                 curEncData.m_rowStat[row].rowQp = qpBase;
                 curEncData.m_rowStat[row].rowQpScale = x265_qp2qScale(qpBase);
@@ -1763,7 +1772,7 @@ void FrameEncoder::processRowEncoder(int intRow, ThreadLocalData& tld)
                 if (curRow.reEncode < 0)
                 {
                     x265_log(m_param, X265_LOG_DEBUG, "POC %d row %d - encode restart required for VBV, to %.2f from %.2f\n",
-                             m_frame->m_poc, row, qpBase, curEncData.m_cuStat[cuAddr].baseQp);
+                             m_frame[layer]->m_poc, row, qpBase, curEncData.m_cuStat[cuAddr].baseQp);
 
                     // prevent the WaveFront::findJob() method from providing new jobs
                     m_vbvResetTriggerRow[curRow.sliceId] = row;
@@ -1984,7 +1993,7 @@ void FrameEncoder::collectDynDataRow(CUData& ctu, FrameStats* rowStats)
     }
 }
 
-void FrameEncoder::collectDynDataFrame()
+void FrameEncoder::collectDynDataFrame(int layer)
 {
     for (uint32_t row = 0; row < m_numRows; row++)
     {
@@ -1993,7 +2002,7 @@ void FrameEncoder::collectDynDataFrame()
             for (uint32_t depth = 0; depth < m_param->maxCUDepth; depth++)
             {
                 int offset = (depth * X265_REFINE_INTER_LEVELS) + refLevel;
-                int curFrameIndex = m_frame->m_encodeOrder - m_top->m_startPoint;
+                int curFrameIndex = m_frame[layer]->m_encodeOrder - m_top->m_startPoint;
                 int index = (curFrameIndex * X265_REFINE_INTER_LEVELS * m_param->maxCUDepth) + offset;
                 if (m_rows[row].rowStats.rowCntDyn[offset])
                 {
@@ -2006,28 +2015,28 @@ void FrameEncoder::collectDynDataFrame()
     }
 }
 
-void FrameEncoder::computeAvgTrainingData()
+void FrameEncoder::computeAvgTrainingData(int layer)
 {
-    if (m_frame->m_lowres.bScenecut || m_frame->m_lowres.bKeyframe)
+    if (m_frame[layer]->m_lowres.bScenecut || m_frame[layer]->m_lowres.bKeyframe)
     {
-        m_top->m_startPoint = m_frame->m_encodeOrder;
+        m_top->m_startPoint = m_frame[layer]->m_encodeOrder;
         int size = (m_param->keyframeMax + m_param->lookaheadDepth) * m_param->maxCUDepth * X265_REFINE_INTER_LEVELS;
         memset(m_top->m_variance, 0, size * sizeof(uint64_t));
         memset(m_top->m_rdCost, 0, size * sizeof(uint64_t));
         memset(m_top->m_trainingCount, 0, size * sizeof(uint32_t));
     }
-    if (m_frame->m_encodeOrder - m_top->m_startPoint < 2 * m_param->frameNumThreads)
-        m_frame->m_classifyFrame = false;
+    if (m_frame[layer]->m_encodeOrder - m_top->m_startPoint < 2 * m_param->frameNumThreads)
+        m_frame[layer]->m_classifyFrame = false;
     else
-        m_frame->m_classifyFrame = true;
+        m_frame[layer]->m_classifyFrame = true;
 
     int size = m_param->maxCUDepth * X265_REFINE_INTER_LEVELS;
-    memset(m_frame->m_classifyRd, 0, size * sizeof(uint64_t));
-    memset(m_frame->m_classifyVariance, 0, size * sizeof(uint64_t));
-    memset(m_frame->m_classifyCount, 0, size * sizeof(uint32_t));
-    if (m_frame->m_classifyFrame)
+    memset(m_frame[layer]->m_classifyRd, 0, size * sizeof(uint64_t));
+    memset(m_frame[layer]->m_classifyVariance, 0, size * sizeof(uint64_t));
+    memset(m_frame[layer]->m_classifyCount, 0, size * sizeof(uint32_t));
+    if (m_frame[layer]->m_classifyFrame)
     {
-        uint32_t limit = m_frame->m_encodeOrder - m_top->m_startPoint - m_param->frameNumThreads;
+        uint32_t limit = m_frame[layer]->m_encodeOrder - m_top->m_startPoint - m_param->frameNumThreads;
         for (uint32_t i = 1; i < limit; i++)
         {
             for (uint32_t j = 0; j < X265_REFINE_INTER_LEVELS; j++)
@@ -2038,15 +2047,15 @@ void FrameEncoder::computeAvgTrainingData()
                     int index = (i* X265_REFINE_INTER_LEVELS * m_param->maxCUDepth) + offset;
                     if (m_top->m_trainingCount[index])
                     {
-                        m_frame->m_classifyRd[offset] += m_top->m_rdCost[index] / m_top->m_trainingCount[index];
-                        m_frame->m_classifyVariance[offset] += m_top->m_variance[index] / m_top->m_trainingCount[index];
-                        m_frame->m_classifyCount[offset] += m_top->m_trainingCount[index];
+                        m_frame[layer]->m_classifyRd[offset] += m_top->m_rdCost[index] / m_top->m_trainingCount[index];
+                        m_frame[layer]->m_classifyVariance[offset] += m_top->m_variance[index] / m_top->m_trainingCount[index];
+                        m_frame[layer]->m_classifyCount[offset] += m_top->m_trainingCount[index];
                     }
                 }
             }
         }
         /* Calculates the average feature values of historic frames that are being considered for the current frame */
-        int historyCount = m_frame->m_encodeOrder - m_param->frameNumThreads - m_top->m_startPoint - 1;
+        int historyCount = m_frame[layer]->m_encodeOrder - m_param->frameNumThreads - m_top->m_startPoint - 1;
         if (historyCount)
         {
             for (uint32_t j = 0; j < X265_REFINE_INTER_LEVELS; j++)
@@ -2054,8 +2063,8 @@ void FrameEncoder::computeAvgTrainingData()
                 for (uint32_t depth = 0; depth < m_param->maxCUDepth; depth++)
                 {
                     int offset = (depth * X265_REFINE_INTER_LEVELS) + j;
-                    m_frame->m_classifyRd[offset] /= historyCount;
-                    m_frame->m_classifyVariance[offset] /= historyCount;
+                    m_frame[layer]->m_classifyRd[offset] /= historyCount;
+                    m_frame[layer]->m_classifyVariance[offset] /= historyCount;
                 }
             }
         }
@@ -2260,13 +2269,13 @@ void FrameEncoder::vmafFrameLevelScore()
 
 Frame *FrameEncoder::getEncodedPicture(NALList& output)
 {
-    if (m_frame)
+    if (m_frame[0])
     {
         /* block here until worker thread completes */
         m_done.wait();
 
-        Frame *ret = m_frame;
-        m_frame = NULL;
+        Frame *ret = m_frame[0];
+        m_frame[0] = NULL;
         output.takeContents(m_nalList);
         m_prevOutputTime = x265_mdate();
         return ret;
diff --git a/source/encoder/frameencoder.h b/source/encoder/frameencoder.h
index 9c5810883..fa4bad60f 100644
--- a/source/encoder/frameencoder.h
+++ b/source/encoder/frameencoder.h
@@ -156,12 +156,12 @@ public:
     void destroy();
 
     /* triggers encode of a new frame by the worker thread */
-    bool startCompressFrame(Frame* curFrame);
+    bool startCompressFrame(Frame* curFrame[MAX_SCALABLE_LAYERS]);
 
     /* blocks until worker thread is done, returns access unit */
     Frame *getEncodedPicture(NALList& list);
 
-    void initDecodedPictureHashSEI(int row, int cuAddr, int height);
+    void initDecodedPictureHashSEI(int row, int cuAddr, int height, int layer);
 
     Event                    m_enable;
     Event                    m_done;
@@ -217,7 +217,7 @@ public:
 
     Encoder*                 m_top;
     x265_param*              m_param;
-    Frame*                   m_frame;
+    Frame*                   m_frame[MAX_SCALABLE_LAYERS];
     NoiseReduction*          m_nr;
     ThreadLocalData*         m_tld; /* for --no-wpp */
     Bitstream*               m_outStreams;
@@ -238,6 +238,8 @@ public:
     TemporalFilter*          m_frameEncTF;
     TemporalFilterRefPicInfo m_mcstfRefList[MAX_MCSTF_TEMPORAL_WINDOW_LENGTH];
 
+    int                      m_sLayerId;
+
     class WeightAnalysis : public BondedTaskGroup
     {
     public:
@@ -258,20 +260,20 @@ protected:
     bool initializeGeoms();
 
     /* analyze / compress frame, can be run in parallel within reference constraints */
-    void compressFrame();
+    void compressFrame(int layer);
 
     /* called by compressFrame to generate final per-row bitstreams */
-    void encodeSlice(uint32_t sliceAddr);
+    void encodeSlice(uint32_t sliceAddr, int layer);
 
     void threadMain();
     int  collectCTUStatistics(const CUData& ctu, FrameStats* frameLog);
     void noiseReductionUpdate();
-    void writeTrailingSEIMessages();
+    void writeTrailingSEIMessages(int layer);
     bool writeToneMapInfo(x265_sei_payload *payload);
 
     /* Called by WaveFront::findJob() */
-    virtual void processRow(int row, int threadId);
-    virtual void processRowEncoder(int row, ThreadLocalData& tld);
+    virtual void processRow(int row, int threadId, int layer);
+    virtual void processRowEncoder(int row, ThreadLocalData& tld, int layer);
 
     void enqueueRowEncoder(int row) { WaveFront::enqueueRow(row * 2 + 0); }
     void enqueueRowFilter(int row)  { WaveFront::enqueueRow(row * 2 + 1); }
@@ -280,8 +282,8 @@ protected:
 #if ENABLE_LIBVMAF
     void vmafFrameLevelScore();
 #endif
-    void collectDynDataFrame();
-    void computeAvgTrainingData();
+    void collectDynDataFrame(int layer);
+    void computeAvgTrainingData(int layer);
     void collectDynDataRow(CUData& ctu, FrameStats* rowStats);    
     void readModel(FilmGrainCharacteristics* m_filmGrain, FILE* filmgrain);
 };
diff --git a/source/encoder/framefilter.cpp b/source/encoder/framefilter.cpp
index 8a09f5b51..6b8ef03ac 100644
--- a/source/encoder/framefilter.cpp
+++ b/source/encoder/framefilter.cpp
@@ -561,7 +561,7 @@ void FrameFilter::ParallelFilter::processTasks(int /*workerThreadId*/)
     }
 }
 
-void FrameFilter::processRow(int row)
+void FrameFilter::processRow(int row, int layer)
 {
     ProfileScopeEvent(filterCTURow);
 
@@ -572,7 +572,7 @@ void FrameFilter::processRow(int row)
 
     if (!m_param->bEnableLoopFilter && !m_useSao)
     {
-        processPostRow(row);
+        processPostRow(row, layer);
         return;
     }
     FrameData& encData = *m_frame->m_encData;
@@ -616,7 +616,7 @@ void FrameFilter::processRow(int row)
 
     // this row of CTUs has been encoded
     if (!ctu->m_bFirstRowInSlice)
-        processPostRow(row - 1);
+        processPostRow(row - 1, layer);
 
     // NOTE: slices parallelism will be execute out-of-order
     int numRowFinished = 0;
@@ -648,10 +648,10 @@ void FrameFilter::processRow(int row)
     }
 
     if (ctu->m_bLastRowInSlice)
-        processPostRow(row);
+        processPostRow(row, layer);
 }
 
-void FrameFilter::processPostRow(int row)
+void FrameFilter::processPostRow(int row, int layer)
 {
     PicYuv *reconPic = m_frame->m_reconPic;
     const uint32_t numCols = m_frame->m_encData->m_slice->m_sps->numCuInWidth;
@@ -713,7 +713,7 @@ void FrameFilter::processPostRow(int row)
     if (m_param->maxSlices == 1)
     {
         uint32_t height = m_parallelFilter[row].getCUHeight();
-        m_frameEncoder->initDecodedPictureHashSEI(row, cuAddr, height);
+        m_frameEncoder->initDecodedPictureHashSEI(row, cuAddr, height, layer);
     } // end of (m_param->maxSlices == 1)
 
     if (ATOMIC_INC(&m_frameEncoder->m_completionCount) == 2 * (int)m_frameEncoder->m_numRows)
diff --git a/source/encoder/framefilter.h b/source/encoder/framefilter.h
index cb0a5038d..b84b7094c 100644
--- a/source/encoder/framefilter.h
+++ b/source/encoder/framefilter.h
@@ -128,8 +128,8 @@ public:
 
     void start(Frame *pic, Entropy& initState);
 
-    void processRow(int row);
-    void processPostRow(int row);
+    void processRow(int row, int layer);
+    void processPostRow(int row, int layer);
     void computeMEIntegral(int row);
 };
 }
diff --git a/source/encoder/nal.cpp b/source/encoder/nal.cpp
index 3e41c5111..093c91016 100644
--- a/source/encoder/nal.cpp
+++ b/source/encoder/nal.cpp
@@ -57,7 +57,7 @@ void NALList::takeContents(NALList& other)
     other.m_buffer = X265_MALLOC(uint8_t, m_allocSize);
 }
 
-void NALList::serialize(NalUnitType nalUnitType, const Bitstream& bs, uint8_t temporalID)
+void NALList::serialize(NalUnitType nalUnitType, const Bitstream& bs, int layerId, uint8_t temporalID)
 {
     static const char startCodePrefix[] = { 0, 0, 0, 1 };
 
@@ -114,7 +114,7 @@ void NALList::serialize(NalUnitType nalUnitType, const Bitstream& bs, uint8_t te
      * nuh_reserved_zero_6bits  6-bits
      * nuh_temporal_id_plus1    3-bits */
     out[bytes++] = (uint8_t)nalUnitType << 1;
-    out[bytes++] = temporalID;
+    out[bytes++] = (layerId << 3) | (temporalID);
 
     /* 7.4.1 ...
      * Within the NAL unit, the following three-byte sequences shall not occur at
diff --git a/source/encoder/nal.h b/source/encoder/nal.h
index 2443413ce..03752cdd6 100644
--- a/source/encoder/nal.h
+++ b/source/encoder/nal.h
@@ -56,7 +56,7 @@ public:
 
     void takeContents(NALList& other);
 
-    void serialize(NalUnitType nalUnitType, const Bitstream& bs, uint8_t temporalID = 1);
+    void serialize(NalUnitType nalUnitType, const Bitstream& bs, int layerId = 0, uint8_t temporalID = 1);
 
     uint32_t serializeSubstreams(uint32_t* streamSizeBytes, uint32_t streamCount, const Bitstream* streams);
 };
diff --git a/source/encoder/sei.cpp b/source/encoder/sei.cpp
index 3002d11ad..6e4b805e5 100644
--- a/source/encoder/sei.cpp
+++ b/source/encoder/sei.cpp
@@ -36,7 +36,7 @@ const uint8_t SEIuserDataUnregistered::m_uuid_iso_iec_11578[16] = {
 
 /* marshal a single SEI message sei, storing the marshalled representation
 * in bitstream bs */
-void SEI::writeSEImessages(Bitstream& bs, const SPS& sps, NalUnitType nalUnitType, NALList& list, int isNested)
+void SEI::writeSEImessages(Bitstream& bs, const SPS& sps, NalUnitType nalUnitType, NALList& list, int isNested, int layer)
 {
     if (!isNested)
         bs.resetBits();
@@ -68,7 +68,7 @@ void SEI::writeSEImessages(Bitstream& bs, const SPS& sps, NalUnitType nalUnitTyp
     {
         if (nalUnitType != NAL_UNIT_UNSPECIFIED)
             bs.writeByteAlignment();
-        list.serialize(nalUnitType, bs, (1 + (nalUnitType == NAL_UNIT_CODED_SLICE_TSA_N)));
+        list.serialize(nalUnitType, bs, layer, (1 + (nalUnitType == NAL_UNIT_CODED_SLICE_TSA_N)));
     }
 }
 
diff --git a/source/encoder/sei.h b/source/encoder/sei.h
index 03e210639..dd4c2b6b8 100644
--- a/source/encoder/sei.h
+++ b/source/encoder/sei.h
@@ -38,7 +38,7 @@ class SEI : public SyntaxElementWriter
 public:
     /* SEI users call writeSEImessages() to marshal an SEI to a bitstream.
     * The writeSEImessages() method calls writeSEI() which encodes the header */
-    void writeSEImessages(Bitstream& bs, const SPS& sps, NalUnitType nalUnitType, NALList& list, int isNested);
+    void writeSEImessages(Bitstream& bs, const SPS& sps, NalUnitType nalUnitType, NALList& list, int isNested, int layerId = 0);
     void setSize(uint32_t size);
     static char* base64Decode(char encodedString[], int base64EncodeLength);
     virtual ~SEI() {}
-- 
2.40.1

